# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import collections

from oslo_config import cfg

from octavia.common import constants
from octavia.tests.common import sample_certs

CONF = cfg.CONF


def sample_amphora_tuple(id='sample_amphora_id_1', lb_network_ip='10.0.1.1',
                         vrrp_ip='10.1.1.1', ha_ip='192.168.10.1',
                         vrrp_port_id='1234', ha_port_id='1234', role=None,
                         status='ACTIVE', vrrp_interface=None,
                         vrrp_priority=None, api_version='0.5'):
    in_amphora = collections.namedtuple(
        'amphora', 'id, lb_network_ip, vrrp_ip, ha_ip, vrrp_port_id, '
                   'ha_port_id, role, status, vrrp_interface,'
                   'vrrp_priority, api_version')
    return in_amphora(
        id=id,
        lb_network_ip=lb_network_ip,
        vrrp_ip=vrrp_ip,
        ha_ip=ha_ip,
        vrrp_port_id=vrrp_port_id,
        ha_port_id=ha_port_id,
        role=role,
        status=status,
        vrrp_interface=vrrp_interface,
        vrrp_priority=vrrp_priority,
        api_version=api_version)


RET_PERSISTENCE = {
    'type': 'HTTP_COOKIE',
    'cookie_name': None}

RET_MONITOR_1 = {
    'id': 'sample_monitor_id_1',
    'type': 'HTTP',
    'delay': 30,
    'timeout': 31,
    'fall_threshold': 3,
    'rise_threshold': 2,
    'http_method': 'GET',
    'url_path': '/index.html',
    'expected_codes': '418',
    'enabled': True,
    'http_version': 1.0,
    'domain_name': None}

RET_MONITOR_2 = {
    'id': 'sample_monitor_id_2',
    'type': 'HTTP',
    'delay': 30,
    'timeout': 31,
    'fall_threshold': 3,
    'rise_threshold': 2,
    'http_method': 'GET',
    'url_path': '/healthmon.html',
    'expected_codes': '418',
    'enabled': True,
    'http_version': 1.0,
    'domain_name': None}

RET_MEMBER_1 = {
    'id': 'sample_member_id_1',
    'address': '10.0.0.99',
    'protocol_port': 82,
    'weight': 13,
    'subnet_id': '10.0.0.1/24',
    'enabled': True,
    'operating_status': 'ACTIVE',
    'monitor_address': None,
    'monitor_port': None,
    'backup': False}

RET_MEMBER_2 = {
    'id': 'sample_member_id_2',
    'address': '10.0.0.98',
    'protocol_port': 82,
    'weight': 13,
    'subnet_id': '10.0.0.1/24',
    'enabled': True,
    'operating_status': 'ACTIVE',
    'monitor_address': None,
    'monitor_port': None,
    'backup': False}

RET_MEMBER_3 = {
    'id': 'sample_member_id_3',
    'address': '10.0.0.97',
    'protocol_port': 82,
    'weight': 13,
    'subnet_id': '10.0.0.1/24',
    'enabled': True,
    'operating_status': 'ACTIVE',
    'monitor_address': None,
    'monitor_port': None,
    'backup': False}

RET_POOL_1 = {
    'id': 'sample_pool_id_1',
    'protocol': 'http',
    'lb_algorithm': 'roundrobin',
    'members': [RET_MEMBER_1, RET_MEMBER_2],
    'health_monitor': RET_MONITOR_1,
    'session_persistence': RET_PERSISTENCE,
    'enabled': True,
    'operating_status': 'ACTIVE',
    'stick_size': '10k',
    constants.HTTP_REUSE: False,
    'ca_tls_path': '',
    'crl_path': '',
    'tls_enabled': False}

RET_POOL_2 = {
    'id': 'sample_pool_id_2',
    'protocol': 'http',
    'lb_algorithm': 'roundrobin',
    'members': [RET_MEMBER_3],
    'health_monitor': RET_MONITOR_2,
    'session_persistence': RET_PERSISTENCE,
    'enabled': True,
    'operating_status': 'ACTIVE',
    'stick_size': '10k',
    constants.HTTP_REUSE: False,
    'ca_tls_path': '',
    'crl_path': '',
    'tls_enabled': False}


RET_DEF_TLS_CONT = {'id': 'cont_id_1', 'allencompassingpem': 'imapem',
                    'primary_cn': 'FakeCn'}
RET_SNI_CONT_1 = {'id': 'cont_id_2', 'allencompassingpem': 'imapem2',
                  'primary_cn': 'FakeCn'}
RET_SNI_CONT_2 = {'id': 'cont_id_3', 'allencompassingpem': 'imapem3',
                  'primary_cn': 'FakeCn2'}

RET_L7RULE_1 = {
    'id': 'sample_l7rule_id_1',
    'type': constants.L7RULE_TYPE_PATH,
    'compare_type': constants.L7RULE_COMPARE_TYPE_STARTS_WITH,
    'key': None,
    'value': '/api',
    'invert': False,
    'enabled': True}

RET_L7RULE_2 = {
    'id': 'sample_l7rule_id_2',
    'type': constants.L7RULE_TYPE_HEADER,
    'compare_type': constants.L7RULE_COMPARE_TYPE_CONTAINS,
    'key': 'Some-header',
    'value': 'This\\ string\\\\\\ with\\ stuff',
    'invert': True,
    'enabled': True}

RET_L7RULE_3 = {
    'id': 'sample_l7rule_id_3',
    'type': constants.L7RULE_TYPE_COOKIE,
    'compare_type': constants.L7RULE_COMPARE_TYPE_REGEX,
    'key': 'some-cookie',
    'value': 'this.*|that',
    'invert': False,
    'enabled': True}

RET_L7RULE_4 = {
    'id': 'sample_l7rule_id_4',
    'type': constants.L7RULE_TYPE_FILE_TYPE,
    'compare_type': constants.L7RULE_COMPARE_TYPE_EQUAL_TO,
    'key': None,
    'value': 'jpg',
    'invert': False,
    'enabled': True}

RET_L7RULE_5 = {
    'id': 'sample_l7rule_id_5',
    'type': constants.L7RULE_TYPE_HOST_NAME,
    'compare_type': constants.L7RULE_COMPARE_TYPE_ENDS_WITH,
    'key': None,
    'value': '.example.com',
    'invert': False,
    'enabled': True}

RET_L7RULE_6 = {
    'id': 'sample_l7rule_id_6',
    'type': constants.L7RULE_TYPE_HOST_NAME,
    'compare_type': constants.L7RULE_COMPARE_TYPE_ENDS_WITH,
    'key': None,
    'value': '.example.com',
    'invert': False,
    'enabled': False}

RET_L7POLICY_1 = {
    'id': 'sample_l7policy_id_1',
    'action': constants.L7POLICY_ACTION_REDIRECT_TO_POOL,
    'redirect_pool': RET_POOL_2,
    'redirect_url': None,
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [RET_L7RULE_1],
    'redirect_http_code': None}

RET_L7POLICY_2 = {
    'id': 'sample_l7policy_id_2',
    'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
    'redirect_pool': None,
    'redirect_url': 'http://www.example.com',
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [RET_L7RULE_2, RET_L7RULE_3],
    'redirect_http_code': 302}

RET_L7POLICY_3 = {
    'id': 'sample_l7policy_id_3',
    'action': constants.L7POLICY_ACTION_REJECT,
    'redirect_pool': None,
    'redirect_url': None,
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [RET_L7RULE_4, RET_L7RULE_5],
    'redirect_http_code': None}

RET_L7POLICY_4 = {
    'id': 'sample_l7policy_id_4',
    'action': constants.L7POLICY_ACTION_REJECT,
    'redirect_pool': None,
    'redirect_url': None,
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [],
    'redirect_http_code': None}

RET_L7POLICY_5 = {
    'id': 'sample_l7policy_id_5',
    'action': constants.L7POLICY_ACTION_REJECT,
    'redirect_pool': None,
    'redirect_url': None,
    'redirect_prefix': None,
    'enabled': False,
    'l7rules': [RET_L7RULE_5],
    'redirect_http_code': None}

RET_L7POLICY_6 = {
    'id': 'sample_l7policy_id_6',
    'action': constants.L7POLICY_ACTION_REJECT,
    'redirect_pool': None,
    'redirect_url': None,
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [],
    'redirect_http_code': None}

RET_L7POLICY_7 = {
    'id': 'sample_l7policy_id_7',
    'action': constants.L7POLICY_ACTION_REDIRECT_PREFIX,
    'redirect_pool': None,
    'redirect_url': None,
    'redirect_prefix': 'https://example.com',
    'enabled': True,
    'l7rules': [RET_L7RULE_2, RET_L7RULE_3],
    'redirect_http_code': 302}

RET_L7POLICY_8 = {
    'id': 'sample_l7policy_id_8',
    'action': constants.L7POLICY_ACTION_REDIRECT_TO_URL,
    'redirect_pool': None,
    'redirect_url': 'http://www.example.com',
    'redirect_prefix': None,
    'enabled': True,
    'l7rules': [RET_L7RULE_2, RET_L7RULE_3],
    'redirect_http_code': None}

RET_LISTENER = {
    'id': 'sample_listener_id_1',
    'protocol_port': '80',
    'protocol': 'HTTP',
    'protocol_mode': 'http',
    'default_pool': RET_POOL_1,
    'connection_limit': constants.HAPROXY_MAX_MAXCONN,
    'amphorae': [sample_amphora_tuple()],
    'peer_port': 1024,
    'topology': 'SINGLE',
    'user_log_format': '12345\\ sample_loadbalancer_id_1\\ %f\\ %ci\\ %cp\\ '
                       '%t\\ %{+Q}r\\ %ST\\ %B\\ %U\\ %[ssl_c_verify]\\ '
                       '%{+Q}[ssl_c_s_dn]\\ %b\\ %s\\ %Tt\\ %tsc',
    'pools': [RET_POOL_1],
    'l7policies': [],
    'enabled': True,
    'insert_headers': {},
    'timeout_client_data': 50000,
    'timeout_member_connect': 5000,
    'timeout_member_data': 50000,
    'timeout_tcp_inspect': 0,
}

RET_LISTENER_L7 = {
    'id': 'sample_listener_id_1',
    'protocol_port': '80',
    'protocol': 'HTTP',
    'protocol_mode': 'http',
    'default_pool': RET_POOL_1,
    'connection_limit': constants.HAPROXY_MAX_MAXCONN,
    'amphorae': [sample_amphora_tuple()],
    'peer_port': 1024,
    'topology': 'SINGLE',
    'user_log_format': '12345\\ sample_loadbalancer_id_1\\ %f\\ %ci\\ %cp\\ '
                       '%t\\ %{+Q}r\\ %ST\\ %B\\ %U\\ %[ssl_c_verify]\\ '
                       '%{+Q}[ssl_c_s_dn]\\ %b\\ %s\\ %Tt\\ %tsc',
    'pools': [RET_POOL_1, RET_POOL_2],
    'l7policies': [RET_L7POLICY_1, RET_L7POLICY_2, RET_L7POLICY_3,
                   RET_L7POLICY_4, RET_L7POLICY_5, RET_L7POLICY_6,
                   RET_L7POLICY_7],
    'enabled': True,
    'insert_headers': {},
    'timeout_client_data': 50000,
    'timeout_member_connect': 5000,
    'timeout_member_data': 50000,
    'timeout_tcp_inspect': 0,
}

RET_LISTENER_TLS = {
    'id': 'sample_listener_id_1',
    'protocol_port': '443',
    'protocol': 'TERMINATED_HTTPS',
    'protocol_mode': 'http',
    'default_pool': RET_POOL_1,
    'connection_limit': constants.HAPROXY_MAX_MAXCONN,
    'tls_certificate_id': 'cont_id_1',
    'default_tls_path': '/etc/ssl/sample_loadbalancer_id_1/fakeCN.pem',
    'default_tls_container': RET_DEF_TLS_CONT,
    'pools': [RET_POOL_1],
    'l7policies': [],
    'enabled': True,
    'insert_headers': {}}

RET_LISTENER_TLS_SNI = {
    'id': 'sample_listener_id_1',
    'protocol_port': '443',
    'protocol': 'TERMINATED_HTTPS',
    'default_pool': RET_POOL_1,
    'connection_limit': constants.HAPROXY_MAX_MAXCONN,
    'tls_certificate_id': 'cont_id_1',
    'default_tls_path': '/etc/ssl/sample_loadbalancer_id_1/fakeCN.pem',
    'default_tls_container': RET_DEF_TLS_CONT,
    'crt_dir': '/v2/sample_loadbalancer_id_1',
    'sni_container_ids': ['cont_id_2', 'cont_id_3'],
    'sni_containers': [RET_SNI_CONT_1, RET_SNI_CONT_2],
    'pools': [RET_POOL_1],
    'l7policies': [],
    'enabled': True,
    'insert_headers': {}}

RET_AMPHORA = {
    'id': 'sample_amphora_id_1',
    'lb_network_ip': '10.0.1.1',
    'vrrp_ip': '10.1.1.1',
    'ha_ip': '192.168.10.1',
    'vrrp_port_id': '1234',
    'ha_port_id': '1234',
    'role': None,
    'status': 'ACTIVE',
    'vrrp_interface': None,
    'vrrp_priority': None}

RET_LB = {
    'host_amphora': RET_AMPHORA,
    'id': 'sample_loadbalancer_id_1',
    'vip_address': '10.0.0.2',
    'listener': RET_LISTENER,
    'topology': 'SINGLE',
    'enabled': True,
    'global_connection_limit': constants.HAPROXY_MAX_MAXCONN}

RET_LB_L7 = {
    'host_amphora': RET_AMPHORA,
    'id': 'sample_loadbalancer_id_1',
    'vip_address': '10.0.0.2',
    'listener': RET_LISTENER_L7,
    'topology': 'SINGLE',
    'enabled': True,
    'global_connection_limit': constants.HAPROXY_MAX_MAXCONN}

UDP_SOURCE_IP_BODY = {
    'type': constants.SESSION_PERSISTENCE_SOURCE_IP,
    'persistence_timeout': 33,
    'persistence_granularity': '255.0.0.0'
}

RET_UDP_HEALTH_MONITOR = {
    'id': 'sample_monitor_id_1',
    'type': constants.HEALTH_MONITOR_UDP_CONNECT,
    'delay': 30,
    'timeout': 31,
    'enabled': True,
    'fall_threshold': 3,
    'check_script_path': (CONF.haproxy_amphora.base_path +
                          '/lvs/check/udp_check.sh')
}

UDP_HEALTH_MONITOR_NO_SCRIPT = {
    'id': 'sample_monitor_id_1',
    'check_script_path': None,
    'delay': 30,
    'enabled': True,
    'fall_threshold': 3,
    'timeout': 31,
    'type': 'UDP'
}

RET_UDP_MEMBER = {
    'id': 'member_id_1',
    'address': '192.0.2.10',
    'protocol_port': 82,
    'weight': 13,
    'enabled': True,
    'monitor_address': None,
    'monitor_port': None
}

RET_UDP_MEMBER_MONITOR_IP_PORT = {
    'id': 'member_id_1',
    'address': '192.0.2.10',
    'protocol_port': 82,
    'weight': 13,
    'enabled': True,
    'monitor_address': '192.168.1.1',
    'monitor_port': 9000
}

UDP_MEMBER_1 = {
    'id': 'sample_member_id_1',
    'address': '10.0.0.99',
    'enabled': True,
    'protocol_port': 82,
    'weight': 13,
    'monitor_address': None,
    'monitor_port': None
}

UDP_MEMBER_2 = {
    'id': 'sample_member_id_2',
    'address': '10.0.0.98',
    'enabled': True,
    'protocol_port': 82,
    'weight': 13,
    'monitor_address': None,
    'monitor_port': None
}

RET_UDP_POOL = {
    'id': 'sample_pool_id_1',
    'enabled': True,
    'health_monitor': UDP_HEALTH_MONITOR_NO_SCRIPT,
    'lb_algorithm': 'wrr',
    'members': [UDP_MEMBER_1, UDP_MEMBER_2],
    'protocol': 'udp',
    'session_persistence': UDP_SOURCE_IP_BODY
}

RET_UDP_LISTENER = {
    'connection_limit': 98,
    'default_pool': {
        'id': 'sample_pool_id_1',
        'enabled': True,
        'health_monitor': RET_UDP_HEALTH_MONITOR,
        'lb_algorithm': 'rr',
        'members': [UDP_MEMBER_1, UDP_MEMBER_2],
        'protocol': 'udp',
        'session_persistence': UDP_SOURCE_IP_BODY
    },
    'enabled': True,
    'id': 'sample_listener_id_1',
    'protocol_mode': 'udp',
    'protocol_port': '80'
}


def sample_loadbalancer_tuple(proto=None, monitor=True, persistence=True,
                              persistence_type=None, tls=False, sni=False,
                              topology=None, l7=False, enabled=True):
    proto = 'HTTP' if proto is None else proto
    topology = 'SINGLE' if topology is None else topology
    in_lb = collections.namedtuple(
        'load_balancer', 'id, name, protocol, vip, listeners, amphorae,'
        ' enabled')
    return in_lb(
        id='sample_loadbalancer_id_1',
        name='test-lb',
        protocol=proto,
        vip=sample_vip_tuple(),
        topology=topology,
        listeners=[sample_listener_tuple(proto=proto, monitor=monitor,
                                         persistence=persistence,
                                         persistence_type=persistence_type,
                                         tls=tls,
                                         sni=sni,
                                         l7=l7,
                                         enabled=enabled)],
        enabled=enabled
    )


def sample_listener_loadbalancer_tuple(proto=None, topology=None,
                                       enabled=True):
    proto = 'HTTP' if proto is None else proto
    if topology and topology in ['ACTIVE_STANDBY', 'ACTIVE_ACTIVE']:
        more_amp = True
    else:
        more_amp = False
        topology = constants.TOPOLOGY_SINGLE
    in_lb = collections.namedtuple(
        'load_balancer', 'id, name, protocol, vip, amphorae, topology, '
        'listeners, enabled, project_id')
    return in_lb(
        id='sample_loadbalancer_id_1',
        name='test-lb',
        protocol=proto,
        vip=sample_vip_tuple(),
        amphorae=[sample_amphora_tuple(role=constants.ROLE_MASTER),
                  sample_amphora_tuple(
                      id='sample_amphora_id_2',
                      lb_network_ip='10.0.1.2',
                      vrrp_ip='10.1.1.2',
                      role=constants.ROLE_BACKUP)]
        if more_amp else [sample_amphora_tuple()],
        topology=topology,
        listeners=[],
        enabled=enabled,
        project_id='12345'
    )


def sample_lb_with_udp_listener_tuple(
        proto=None, topology=None, enabled=True, pools=None):
    proto = 'HTTP' if proto is None else proto
    if topology and topology in ['ACTIVE_STANDBY', 'ACTIVE_ACTIVE']:
        more_amp = True
    else:
        more_amp = False
        topology = constants.TOPOLOGY_SINGLE
    listeners = [sample_listener_tuple(
        proto=constants.PROTOCOL_UDP,
        persistence_type=constants.SESSION_PERSISTENCE_SOURCE_IP,
        persistence_timeout=33,
        persistence_granularity='255.255.0.0',
        monitor_proto=constants.HEALTH_MONITOR_UDP_CONNECT)]

    in_lb = collections.namedtuple(
        'load_balancer', 'id, name, protocol, vip, amphorae, topology, '
        'pools, enabled, project_id, listeners')
    return in_lb(
        id='sample_loadbalancer_id_1',
        name='test-lb',
        protocol=proto,
        vip=sample_vip_tuple(),
        amphorae=[sample_amphora_tuple(role=constants.ROLE_MASTER),
                  sample_amphora_tuple(
                      id='sample_amphora_id_2',
                      lb_network_ip='10.0.1.2',
                      vrrp_ip='10.1.1.2',
                      role=constants.ROLE_BACKUP)]
        if more_amp else [sample_amphora_tuple()],
        topology=topology,
        listeners=listeners,
        pools=pools or [],
        enabled=enabled,
        project_id='12345'
    )


def sample_vrrp_group_tuple():
    in_vrrp_group = collections.namedtuple(
        'vrrp_group', 'load_balancer_id, vrrp_auth_type, vrrp_auth_pass, '
                      'advert_int, smtp_server, smtp_connect_timeout, '
                      'vrrp_group_name')
    return in_vrrp_group(
        vrrp_group_name='sample_loadbalancer_id_1',
        load_balancer_id='sample_loadbalancer_id_1',
        vrrp_auth_type='PASS',
        vrrp_auth_pass='123',
        advert_int='1',
        smtp_server='',
        smtp_connect_timeout='')


def sample_vip_tuple():
    vip = collections.namedtuple('vip', 'ip_address')
    return vip(ip_address='10.0.0.2')


def sample_listener_tuple(proto=None, monitor=True, alloc_default_pool=True,
                          persistence=True, persistence_type=None,
                          persistence_cookie=None, persistence_timeout=None,
                          persistence_granularity=None,
                          tls=False, sni=False, peer_port=None, topology=None,
                          l7=False, enabled=True, insert_headers=None,
                          be_proto=None, monitor_ip_port=False,
                          monitor_proto=None, monitor_expected_codes=None,
                          backup_member=False, disabled_member=False,
                          connection_limit=-1,
                          timeout_client_data=50000,
                          timeout_member_connect=5000,
                          timeout_member_data=50000,
                          timeout_tcp_inspect=0,
                          client_ca_cert=False, client_crl_cert=False,
                          ssl_type_l7=False, pool_cert=False,
                          pool_ca_cert=False, pool_crl=False,
                          tls_enabled=False, hm_host_http_check=False,
                          id='sample_listener_id_1', recursive_nest=False,
                          provisioning_status=constants.ACTIVE):
    proto = 'HTTP' if proto is None else proto
    if be_proto is None:
        be_proto = 'HTTP' if proto == 'TERMINATED_HTTPS' else proto
    topology = 'SINGLE' if topology is None else topology
    port = '443' if proto in ['HTTPS', 'TERMINATED_HTTPS'] else '80'
    peer_port = 1024 if peer_port is None else peer_port
    insert_headers = insert_headers or {}
    in_listener = collections.namedtuple(
        'listener', 'id, project_id, protocol_port, protocol, default_pool, '
                    'connection_limit, tls_certificate_id, '
                    'sni_container_ids, default_tls_container, '
                    'sni_containers, load_balancer, peer_port, pools, '
                    'l7policies, enabled, insert_headers, timeout_client_data,'
                    'timeout_member_connect, timeout_member_data, '
                    'timeout_tcp_inspect, client_ca_tls_certificate_id, '
                    'client_ca_tls_certificate, client_authentication, '
                    'client_crl_container_id, provisioning_status')
    if l7:
        pools = [
            sample_pool_tuple(
                proto=be_proto, monitor=monitor, persistence=persistence,
                persistence_type=persistence_type,
                persistence_cookie=persistence_cookie,
                monitor_ip_port=monitor_ip_port, monitor_proto=monitor_proto,
                pool_cert=pool_cert, pool_ca_cert=pool_ca_cert,
                pool_crl=pool_crl, tls_enabled=tls_enabled,
                hm_host_http_check=hm_host_http_check),
            sample_pool_tuple(
                proto=be_proto, monitor=monitor, persistence=persistence,
                persistence_type=persistence_type,
                persistence_cookie=persistence_cookie, sample_pool=2,
                monitor_ip_port=monitor_ip_port, monitor_proto=monitor_proto,
                pool_cert=pool_cert, pool_ca_cert=pool_ca_cert,
                pool_crl=pool_crl, tls_enabled=tls_enabled,
                hm_host_http_check=hm_host_http_check)]
        l7policies = [
            sample_l7policy_tuple('sample_l7policy_id_1', sample_policy=1),
            sample_l7policy_tuple('sample_l7policy_id_2', sample_policy=2),
            sample_l7policy_tuple('sample_l7policy_id_3', sample_policy=3),
            sample_l7policy_tuple('sample_l7policy_id_4', sample_policy=4),
            sample_l7policy_tuple('sample_l7policy_id_5', sample_policy=5),
            sample_l7policy_tuple('sample_l7policy_id_6', sample_policy=6),
            sample_l7policy_tuple('sample_l7policy_id_7', sample_policy=7)]
        if ssl_type_l7:
            l7policies.append(sample_l7policy_tuple(
                'sample_l7policy_id_8', sample_policy=8))
    else:
        pools = [
            sample_pool_tuple(
                proto=be_proto, monitor=monitor, persistence=persistence,
                persistence_type=persistence_type,
                persistence_cookie=persistence_cookie,
                monitor_ip_port=monitor_ip_port, monitor_proto=monitor_proto,
                backup_member=backup_member, disabled_member=disabled_member,
                pool_cert=pool_cert, pool_ca_cert=pool_ca_cert,
                pool_crl=pool_crl, tls_enabled=tls_enabled,
                hm_host_http_check=hm_host_http_check)]
        l7policies = []
    listener = in_listener(
        id=id,
        project_id='12345',
        protocol_port=port,
        protocol=proto,
        load_balancer=sample_listener_loadbalancer_tuple(proto=proto,
                                                         topology=topology),
        peer_port=peer_port,
        default_pool=sample_pool_tuple(
            proto=be_proto, monitor=monitor, persistence=persistence,
            persistence_type=persistence_type,
            persistence_cookie=persistence_cookie,
            persistence_timeout=persistence_timeout,
            persistence_granularity=persistence_granularity,
            monitor_ip_port=monitor_ip_port,
            monitor_proto=monitor_proto,
            monitor_expected_codes=monitor_expected_codes,
            pool_cert=pool_cert,
            pool_ca_cert=pool_ca_cert,
            pool_crl=pool_crl,
            tls_enabled=tls_enabled,
            hm_host_http_check=hm_host_http_check
        ) if alloc_default_pool else '',
        connection_limit=connection_limit,
        tls_certificate_id='cont_id_1' if tls else '',
        sni_container_ids=['cont_id_2', 'cont_id_3'] if sni else [],
        default_tls_container=sample_tls_container_tuple(
            id='cont_id_1', certificate=sample_certs.X509_CERT,
            private_key=sample_certs.X509_CERT_KEY,
            intermediates=sample_certs.X509_IMDS_LIST,
            primary_cn=sample_certs.X509_CERT_CN
        ) if tls else '',
        sni_containers=[
            sample_tls_sni_container_tuple(
                tls_container_id='cont_id_2',
                tls_container=sample_tls_container_tuple(
                    id='cont_id_2', certificate=sample_certs.X509_CERT_2,
                    private_key=sample_certs.X509_CERT_KEY_2,
                    intermediates=sample_certs.X509_IMDS_LIST,
                    primary_cn=sample_certs.X509_CERT_CN_2)),
            sample_tls_sni_container_tuple(
                tls_container_id='cont_id_3',
                tls_container=sample_tls_container_tuple(
                    id='cont_id_3', certificate=sample_certs.X509_CERT_3,
                    private_key=sample_certs.X509_CERT_KEY_3,
                    intermediates=sample_certs.X509_IMDS_LIST,
                    primary_cn=sample_certs.X509_CERT_CN_3))]
        if sni else [],
        pools=pools,
        l7policies=l7policies,
        enabled=enabled,
        insert_headers=insert_headers,
        timeout_client_data=timeout_client_data,
        timeout_member_connect=timeout_member_connect,
        timeout_member_data=timeout_member_data,
        timeout_tcp_inspect=timeout_tcp_inspect,
        client_ca_tls_certificate_id='cont_id_ca' if client_ca_cert else '',
        client_ca_tls_certificate=sample_tls_container_tuple(
            id='cont_id_ca', certificate=sample_certs.X509_CA_CERT,
            primary_cn=sample_certs.X509_CA_CERT_CN
        ) if client_ca_cert else '',
        client_authentication=(
            constants.CLIENT_AUTH_MANDATORY if client_ca_cert else
            constants.CLIENT_AUTH_NONE),
        client_crl_container_id='cont_id_crl' if client_crl_cert else '',
        provisioning_status=provisioning_status,
    )
    if recursive_nest:
        listener.load_balancer.listeners.append(listener)
    return listener


def sample_tls_sni_container_tuple(tls_container_id=None, tls_container=None):
    sc = collections.namedtuple('sni_container', 'tls_container_id, '
                                                 'tls_container')
    return sc(tls_container_id=tls_container_id, tls_container=tls_container)


def sample_tls_sni_containers_tuple(tls_container_id=None, tls_container=None):
    sc = collections.namedtuple('sni_containers', 'tls_container_id, '
                                                  'tls_container')
    return [sc(tls_container_id=tls_container_id, tls_container=tls_container)]


def sample_tls_container_tuple(id='cont_id_1', certificate=None,
                               private_key=None, intermediates=None,
                               primary_cn=None):
    sc = collections.namedtuple(
        'tls_container',
        'id, certificate, private_key, intermediates, primary_cn')
    return sc(id=id, certificate=certificate, private_key=private_key,
              intermediates=intermediates or [], primary_cn=primary_cn)


def sample_pool_tuple(proto=None, monitor=True, persistence=True,
                      persistence_type=None, persistence_cookie=None,
                      persistence_timeout=None, persistence_granularity=None,
                      sample_pool=1, monitor_ip_port=False,
                      monitor_proto=None, monitor_expected_codes=None,
                      backup_member=False,
                      disabled_member=False, has_http_reuse=True,
                      pool_cert=False, pool_ca_cert=False, pool_crl=False,
                      tls_enabled=False, hm_host_http_check=False,
                      provisioning_status=constants.ACTIVE):
    proto = 'HTTP' if proto is None else proto
    monitor_proto = proto if monitor_proto is None else monitor_proto
    in_pool = collections.namedtuple(
        'pool', 'id, protocol, lb_algorithm, members, health_monitor, '
                'session_persistence, enabled, operating_status, '
                'tls_certificate_id, ca_tls_certificate_id, '
                'crl_container_id, tls_enabled, provisioning_status, ' +
                constants.HTTP_REUSE)
    if (proto == constants.PROTOCOL_UDP and
            persistence_type == constants.SESSION_PERSISTENCE_SOURCE_IP):
        kwargs = {'persistence_type': persistence_type,
                  'persistence_timeout': persistence_timeout,
                  'persistence_granularity': persistence_granularity}
    else:
        kwargs = {'persistence_type': persistence_type,
                  'persistence_cookie': persistence_cookie}
    persis = sample_session_persistence_tuple(**kwargs)
    mon = None
    if sample_pool == 1:
        id = 'sample_pool_id_1'
        members = [sample_member_tuple('sample_member_id_1', '10.0.0.99',
                                       monitor_ip_port=monitor_ip_port),
                   sample_member_tuple('sample_member_id_2', '10.0.0.98',
                                       monitor_ip_port=monitor_ip_port,
                                       backup=backup_member,
                                       enabled=not disabled_member)]
        if monitor is True:
            mon = sample_health_monitor_tuple(
                proto=monitor_proto,
                host_http_check=hm_host_http_check,
                expected_codes=monitor_expected_codes)
    elif sample_pool == 2:
        id = 'sample_pool_id_2'
        members = [sample_member_tuple('sample_member_id_3', '10.0.0.97',
                                       monitor_ip_port=monitor_ip_port)]
        if monitor is True:
            mon = sample_health_monitor_tuple(
                proto=monitor_proto, sample_hm=2,
                host_http_check=hm_host_http_check,
                expected_codes=monitor_expected_codes)

    return in_pool(
        id=id,
        protocol=proto,
        lb_algorithm='ROUND_ROBIN',
        members=members,
        health_monitor=mon,
        session_persistence=persis if persistence is True else None,
        enabled=True,
        operating_status='ACTIVE', has_http_reuse=has_http_reuse,
        tls_certificate_id='pool_cont_1' if pool_cert else None,
        ca_tls_certificate_id='pool_ca_1' if pool_ca_cert else None,
        crl_container_id='pool_crl' if pool_crl else None,
        tls_enabled=tls_enabled, provisioning_status=provisioning_status)


def sample_member_tuple(id, ip, enabled=True,
                        operating_status=constants.ACTIVE,
                        provisioning_status=constants.ACTIVE,
                        monitor_ip_port=False, backup=False):
    in_member = collections.namedtuple('member',
                                       'id, ip_address, protocol_port, '
                                       'weight, subnet_id, '
                                       'enabled, operating_status, '
                                       'monitor_address, monitor_port, '
                                       'backup, provisioning_status')
    monitor_address = '192.168.1.1' if monitor_ip_port else None
    monitor_port = 9000 if monitor_ip_port else None
    return in_member(
        id=id,
        ip_address=ip,
        protocol_port=82,
        weight=13,
        subnet_id='10.0.0.1/24',
        enabled=enabled,
        operating_status=operating_status,
        monitor_address=monitor_address,
        monitor_port=monitor_port,
        backup=backup, provisioning_status=provisioning_status)


def sample_session_persistence_tuple(persistence_type=None,
                                     persistence_cookie=None,
                                     persistence_timeout=None,
                                     persistence_granularity=None):
    spersistence = collections.namedtuple('SessionPersistence',
                                          'type, cookie_name, '
                                          'persistence_timeout, '
                                          'persistence_granularity')
    pt = 'HTTP_COOKIE' if persistence_type is None else persistence_type
    return spersistence(type=pt,
                        cookie_name=persistence_cookie,
                        persistence_timeout=persistence_timeout,
                        persistence_granularity=persistence_granularity)


def sample_health_monitor_tuple(proto='HTTP', sample_hm=1,
                                host_http_check=False, expected_codes=None,
                                provisioning_status=constants.ACTIVE):
    proto = 'HTTP' if proto == 'TERMINATED_HTTPS' else proto
    monitor = collections.namedtuple(
        'monitor', 'id, type, delay, timeout, fall_threshold, rise_threshold,'
                   'http_method, url_path, expected_codes, enabled, '
                   'check_script_path, http_version, domain_name, '
                   'provisioning_status')

    if sample_hm == 1:
        id = 'sample_monitor_id_1'
        url_path = '/index.html'
    elif sample_hm == 2:
        id = 'sample_monitor_id_2'
        url_path = '/healthmon.html'
    kwargs = {
        'id': id,
        'type': proto,
        'delay': 30,
        'timeout': 31,
        'fall_threshold': 3,
        'rise_threshold': 2,
        'http_method': 'GET',
        'url_path': url_path,
        'expected_codes': '418',
        'enabled': True,
        'provisioning_status': provisioning_status,
    }
    if host_http_check:
        kwargs.update({'http_version': 1.1, 'domain_name': 'testlab.com'})
    else:
        kwargs.update({'http_version': 1.0, 'domain_name': None})
    if expected_codes:
        kwargs.update({'expected_codes': expected_codes})
    if proto == constants.HEALTH_MONITOR_UDP_CONNECT:
        kwargs['check_script_path'] = (CONF.haproxy_amphora.base_path +
                                       'lvs/check/' + 'udp_check.sh')
    else:
        kwargs['check_script_path'] = None
    return monitor(**kwargs)


def sample_l7policy_tuple(id,
                          action=constants.L7POLICY_ACTION_REJECT,
                          redirect_pool=None, redirect_url=None,
                          redirect_prefix=None,
                          enabled=True, redirect_http_code=302,
                          sample_policy=1,
                          provisioning_status=constants.ACTIVE):
    in_l7policy = collections.namedtuple('l7policy',
                                         'id, action, redirect_pool, '
                                         'redirect_url, redirect_prefix, '
                                         'l7rules, enabled, '
                                         'redirect_http_code, '
                                         'provisioning_status')
    l7rules = []
    if sample_policy == 1:
        action = constants.L7POLICY_ACTION_REDIRECT_TO_POOL
        redirect_pool = sample_pool_tuple(sample_pool=2)
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_1')]
    elif sample_policy == 2:
        action = constants.L7POLICY_ACTION_REDIRECT_TO_URL
        redirect_url = 'http://www.example.com'
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_2', sample_rule=2),
                   sample_l7rule_tuple('sample_l7rule_id_3', sample_rule=3)]
    elif sample_policy == 3:
        action = constants.L7POLICY_ACTION_REJECT
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_4', sample_rule=4),
                   sample_l7rule_tuple('sample_l7rule_id_5', sample_rule=5)]
    elif sample_policy == 4:
        action = constants.L7POLICY_ACTION_REJECT
    elif sample_policy == 5:
        action = constants.L7POLICY_ACTION_REJECT
        enabled = False
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_5', sample_rule=5)]
    elif sample_policy == 6:
        action = constants.L7POLICY_ACTION_REJECT
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_6', sample_rule=6)]
    elif sample_policy == 7:
        action = constants.L7POLICY_ACTION_REDIRECT_PREFIX
        redirect_prefix = 'https://example.com'
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_2', sample_rule=2),
                   sample_l7rule_tuple('sample_l7rule_id_3', sample_rule=3)]
    elif sample_policy == 8:
        action = constants.L7POLICY_ACTION_REDIRECT_TO_URL
        redirect_url = 'http://www.ssl-type-l7rule-test.com'
        l7rules = [sample_l7rule_tuple('sample_l7rule_id_7', sample_rule=7),
                   sample_l7rule_tuple('sample_l7rule_id_8', sample_rule=8),
                   sample_l7rule_tuple('sample_l7rule_id_9', sample_rule=9),
                   sample_l7rule_tuple('sample_l7rule_id_10', sample_rule=10),
                   sample_l7rule_tuple('sample_l7rule_id_11', sample_rule=11)]
    return in_l7policy(
        id=id,
        action=action,
        redirect_pool=redirect_pool,
        redirect_url=redirect_url,
        redirect_prefix=redirect_prefix,
        l7rules=l7rules,
        enabled=enabled,
        redirect_http_code=redirect_http_code
        if (action in [constants.L7POLICY_ACTION_REDIRECT_TO_URL,
                       constants.L7POLICY_ACTION_REDIRECT_PREFIX] and
            redirect_http_code) else None,
        provisioning_status=provisioning_status)


def sample_l7rule_tuple(id,
                        type=constants.L7RULE_TYPE_PATH,
                        compare_type=constants.L7RULE_COMPARE_TYPE_STARTS_WITH,
                        key=None, value='/api',
                        invert=False, enabled=True,
                        sample_rule=1, provisioning_status=constants.ACTIVE):
    in_l7rule = collections.namedtuple('l7rule',
                                       'id, type, compare_type, '
                                       'key, value, invert, enabled, '
                                       'provisioning_status')
    if sample_rule == 2:
        type = constants.L7RULE_TYPE_HEADER
        compare_type = constants.L7RULE_COMPARE_TYPE_CONTAINS
        key = 'Some-header'
        value = 'This string\\ with stuff'
        invert = True
        enabled = True
    if sample_rule == 3:
        type = constants.L7RULE_TYPE_COOKIE
        compare_type = constants.L7RULE_COMPARE_TYPE_REGEX
        key = 'some-cookie'
        value = 'this.*|that'
        invert = False
        enabled = True
    if sample_rule == 4:
        type = constants.L7RULE_TYPE_FILE_TYPE
        compare_type = constants.L7RULE_COMPARE_TYPE_EQUAL_TO
        key = None
        value = 'jpg'
        invert = False
        enabled = True
    if sample_rule == 5:
        type = constants.L7RULE_TYPE_HOST_NAME
        compare_type = constants.L7RULE_COMPARE_TYPE_ENDS_WITH
        key = None
        value = '.example.com'
        invert = False
        enabled = True
    if sample_rule == 6:
        type = constants.L7RULE_TYPE_HOST_NAME
        compare_type = constants.L7RULE_COMPARE_TYPE_ENDS_WITH
        key = None
        value = '.example.com'
        invert = False
        enabled = False
    if sample_rule == 7:
        type = constants.L7RULE_TYPE_SSL_CONN_HAS_CERT
        compare_type = constants.L7RULE_COMPARE_TYPE_EQUAL_TO
        key = None
        value = 'tRuE'
        invert = False
        enabled = True
    if sample_rule == 8:
        type = constants.L7RULE_TYPE_SSL_VERIFY_RESULT
        compare_type = constants.L7RULE_COMPARE_TYPE_EQUAL_TO
        key = None
        value = '1'
        invert = True
        enabled = True
    if sample_rule == 9:
        type = constants.L7RULE_TYPE_SSL_DN_FIELD
        compare_type = constants.L7RULE_COMPARE_TYPE_REGEX
        key = 'STREET'
        value = r'^STREET.*NO\.$'
        invert = True
        enabled = True
    if sample_rule == 10:
        type = constants.L7RULE_TYPE_SSL_DN_FIELD
        compare_type = constants.L7RULE_COMPARE_TYPE_STARTS_WITH
        key = 'OU-3'
        value = 'Orgnization Bala'
        invert = True
        enabled = True
    return in_l7rule(
        id=id,
        type=type,
        compare_type=compare_type,
        key=key,
        value=value,
        invert=invert,
        enabled=enabled, provisioning_status=provisioning_status)


def sample_base_expected_config(frontend=None, logging=None, backend=None,
                                peers=None, global_opts=None, defaults=None):
    if frontend is None:
        frontend = ("frontend sample_listener_id_1\n"
                    "    maxconn {maxconn}\n"
                    "    bind 10.0.0.2:80\n"
                    "    mode http\n"
                    "    default_backend sample_pool_id_1\n"
                    "    timeout client 50000\n").format(
            maxconn=constants.HAPROXY_MAX_MAXCONN)
    if logging is None:
        logging = ("    log-format 12345\\ sample_loadbalancer_id_1\\ %f\\ "
                   "%ci\\ %cp\\ %t\\ %{+Q}r\\ %ST\\ %B\\ %U\\ "
                   "%[ssl_c_verify]\\ %{+Q}[ssl_c_s_dn]\\ %b\\ %s\\ %Tt\\ "
                   "%tsc\n\n")
    if backend is None:
        backend = ("backend sample_pool_id_1\n"
                   "    mode http\n"
                   "    balance roundrobin\n"
                   "    cookie SRV insert indirect nocache\n"
                   "    timeout check 31s\n"
                   "    option httpchk GET /index.html HTTP/1.0\\r\\n\n"
                   "    http-check expect rstatus 418\n"
                   "    fullconn {maxconn}\n"
                   "    option allbackups\n"
                   "    timeout connect 5000\n"
                   "    timeout server 50000\n"
                   "    server sample_member_id_1 10.0.0.99:82 weight 13 "
                   "check inter 30s fall 3 rise 2 cookie sample_member_id_1\n"
                   "    server sample_member_id_2 10.0.0.98:82 weight 13 "
                   "check inter 30s fall 3 rise 2 cookie sample_member_id_2\n"
                   "\n").format(maxconn=constants.HAPROXY_MAX_MAXCONN)

    if peers is None:
        peers = "\n\n"
    if global_opts is None:
        global_opts = "    maxconn {maxconn}\n\n".format(
            maxconn=constants.HAPROXY_MAX_MAXCONN)
    if defaults is None:
        defaults = ("defaults\n"
                    "    log global\n"
                    "    retries 3\n"
                    "    option redispatch\n"
                    "    option splice-request\n"
                    "    option splice-response\n"
                    "    option http-keep-alive\n\n")
    return ("# Configuration for loadbalancer sample_loadbalancer_id_1\n"
            "global\n"
            "    daemon\n"
            "    user nobody\n"
            "    log /run/rsyslog/octavia/log local0\n"
            "    log /run/rsyslog/octavia/log local1 notice\n"
            "    stats socket /var/lib/octavia/sample_listener_id_1.sock"
            " mode 0666 level user\n" +
            global_opts + defaults + peers + frontend + logging + backend)



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from octavia_lib.api.drivers import data_models as driver_dm
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_utils import uuidutils

from octavia.api.drivers import utils as driver_utils
from octavia.common import constants
import octavia.common.context
from octavia.common import data_models
from octavia.common import exceptions
from octavia.db import repositories
from octavia.network import base as network_base
from octavia.tests.functional.api.v2 import base


class TestMember(base.BaseAPITest):

    root_tag = 'member'
    root_tag_list = 'members'
    root_tag_links = 'members_links'

    def setUp(self):
        super(TestMember, self).setUp()
        vip_subnet_id = uuidutils.generate_uuid()
        self.lb = self.create_load_balancer(vip_subnet_id)
        self.lb_id = self.lb.get('loadbalancer').get('id')
        self.project_id = self.lb.get('loadbalancer').get('project_id')
        self.set_lb_status(self.lb_id)
        self.listener = self.create_listener(
            constants.PROTOCOL_HTTP, 80,
            lb_id=self.lb_id)
        self.listener_id = self.listener.get('listener').get('id')
        self.set_lb_status(self.lb_id)
        self.pool = self.create_pool(self.lb_id, constants.PROTOCOL_HTTP,
                                     constants.LB_ALGORITHM_ROUND_ROBIN)
        self.pool_id = self.pool.get('pool').get('id')
        self.set_lb_status(self.lb_id)
        self.pool_with_listener = self.create_pool(
            self.lb_id, constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN, listener_id=self.listener_id)
        self.pool_with_listener_id = (
            self.pool_with_listener.get('pool').get('id'))
        self.set_lb_status(self.lb_id)
        self.members_path = self.MEMBERS_PATH.format(
            pool_id=self.pool_id)
        self.member_path = self.members_path + '/{member_id}'
        self.members_path_listener = self.MEMBERS_PATH.format(
            pool_id=self.pool_with_listener_id)
        self.member_path_listener = self.members_path_listener + '/{member_id}'
        self.pool_repo = repositories.PoolRepository()

    def test_get(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        response = self.get(self.member_path.format(
            member_id=api_member.get('id'))).json.get(self.root_tag)
        self.assertEqual(api_member, response)
        self.assertEqual(api_member.get('name'), '')
        self.assertEqual([], api_member['tags'])

    def test_get_authorized(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': self.project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.get(self.member_path.format(
                    member_id=api_member.get('id'))).json.get(self.root_tag)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(api_member, response)
        self.assertEqual(api_member.get('name'), '')

    def test_get_not_authorized(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               uuidutils.generate_uuid()):
            response = self.get(self.member_path.format(
                member_id=api_member.get('id')), status=403).json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, response)

    def test_get_deleted_gives_404(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)

        self.set_object_status(self.member_repo, api_member.get('id'),
                               provisioning_status=constants.DELETED)
        self.get(self.member_path.format(member_id=api_member.get('id')),
                 status=404)

    def test_bad_get(self):
        self.get(self.member_path.format(member_id=uuidutils.generate_uuid()),
                 status=404)

    def test_get_all(self):
        api_m_1 = self.create_member(
            self.pool_id, '192.0.2.1', 80,
            tags=['test_tag1']).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        api_m_2 = self.create_member(
            self.pool_id, '192.0.2.2', 80,
            tags=['test_tag2']).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        # Original objects didn't have the updated operating/provisioning
        # status that exists in the DB.
        for m in [api_m_1, api_m_2]:
            m['operating_status'] = constants.ONLINE
            m['provisioning_status'] = constants.ACTIVE
            m.pop('updated_at')
        response = self.get(self.members_path).json.get(self.root_tag_list)
        self.assertIsInstance(response, list)
        self.assertEqual(2, len(response))
        for m in response:
            m.pop('updated_at')
        for m in [api_m_1, api_m_2]:
            self.assertIn(m, response)

    def test_get_all_hides_deleted(self):
        api_member = self.create_member(
            self.pool_id, '10.0.0.1', 80).get(self.root_tag)

        response = self.get(self.members_path)
        objects = response.json.get(self.root_tag_list)
        self.assertEqual(len(objects), 1)
        self.set_object_status(self.member_repo, api_member.get('id'),
                               provisioning_status=constants.DELETED)
        response = self.get(self.members_path)
        objects = response.json.get(self.root_tag_list)
        self.assertEqual(len(objects), 0)

    def test_get_all_authorized(self):
        api_m_1 = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        api_m_2 = self.create_member(
            self.pool_id, '192.0.2.2', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        # Original objects didn't have the updated operating/provisioning
        # status that exists in the DB.
        for m in [api_m_1, api_m_2]:
            m['operating_status'] = constants.ONLINE
            m['provisioning_status'] = constants.ACTIVE
            m.pop('updated_at')

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': self.project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.get(self.members_path)
                response = response.json.get(self.root_tag_list)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        self.assertIsInstance(response, list)
        self.assertEqual(2, len(response))
        for m in response:
            m.pop('updated_at')
        for m in [api_m_1, api_m_2]:
            self.assertIn(m, response)

    def test_get_all_unscoped_token(self):
        api_m_1 = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        api_m_2 = self.create_member(
            self.pool_id, '192.0.2.2', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        # Original objects didn't have the updated operating/provisioning
        # status that exists in the DB.
        for m in [api_m_1, api_m_2]:
            m['operating_status'] = constants.ONLINE
            m['provisioning_status'] = constants.ACTIVE
            m.pop('updated_at')

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               None):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': None}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                result = self.get(self.members_path, status=403).json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, result)

    def test_get_all_not_authorized(self):
        api_m_1 = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        api_m_2 = self.create_member(
            self.pool_id, '192.0.2.2', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        # Original objects didn't have the updated operating/provisioning
        # status that exists in the DB.
        for m in [api_m_1, api_m_2]:
            m['operating_status'] = constants.ONLINE
            m['provisioning_status'] = constants.ACTIVE
            m.pop('updated_at')

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               uuidutils.generate_uuid()):
            response = self.get(self.members_path, status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, response.json)

    def test_get_all_sorted(self):
        self.create_member(self.pool_id, '192.0.2.1', 80, name='member1')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.2', 80, name='member2')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.3', 80, name='member3')
        self.set_lb_status(self.lb_id)

        response = self.get(self.members_path,
                            params={'sort': 'name:desc'})
        members_desc = response.json.get(self.root_tag_list)
        response = self.get(self.members_path,
                            params={'sort': 'name:asc'})
        members_asc = response.json.get(self.root_tag_list)

        self.assertEqual(3, len(members_desc))
        self.assertEqual(3, len(members_asc))

        member_id_names_desc = [(member.get('id'), member.get('name'))
                                for member in members_desc]
        member_id_names_asc = [(member.get('id'), member.get('name'))
                               for member in members_asc]
        self.assertEqual(member_id_names_asc,
                         list(reversed(member_id_names_desc)))

    def test_get_all_limited(self):
        self.create_member(self.pool_id, '192.0.2.1', 80, name='member1')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.2', 80, name='member2')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.3', 80, name='member3')
        self.set_lb_status(self.lb_id)

        # First two -- should have 'next' link
        first_two = self.get(self.members_path, params={'limit': 2}).json
        objs = first_two[self.root_tag_list]
        links = first_two[self.root_tag_links]
        self.assertEqual(2, len(objs))
        self.assertEqual(1, len(links))
        self.assertEqual('next', links[0]['rel'])

        # Third + off the end -- should have previous link
        third = self.get(self.members_path, params={
            'limit': 2,
            'marker': first_two[self.root_tag_list][1]['id']}).json
        objs = third[self.root_tag_list]
        links = third[self.root_tag_links]
        self.assertEqual(1, len(objs))
        self.assertEqual(1, len(links))
        self.assertEqual('previous', links[0]['rel'])

        # Middle -- should have both links
        middle = self.get(self.members_path, params={
            'limit': 1,
            'marker': first_two[self.root_tag_list][0]['id']}).json
        objs = middle[self.root_tag_list]
        links = middle[self.root_tag_links]
        self.assertEqual(1, len(objs))
        self.assertEqual(2, len(links))
        self.assertItemsEqual(['previous', 'next'],
                              [link['rel'] for link in links])

    def test_get_all_fields_filter(self):
        self.create_member(self.pool_id, '192.0.2.1', 80, name='member1')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.2', 80, name='member2')
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id, '192.0.2.3', 80, name='member3')
        self.set_lb_status(self.lb_id)

        members = self.get(self.members_path, params={
            'fields': ['id', 'address']}).json
        for member in members['members']:
            self.assertIn(u'id', member)
            self.assertIn(u'address', member)
            self.assertNotIn(u'name', member)
            self.assertNotIn(u'monitor_address', member)

    def test_get_one_fields_filter(self):
        member1 = self.create_member(
            self.pool_id, '192.0.2.1', 80, name='member1').get(self.root_tag)
        self.set_lb_status(self.lb_id)

        member = self.get(
            self.member_path.format(member_id=member1.get('id')),
            params={'fields': ['id', 'address']}).json.get(self.root_tag)
        self.assertIn(u'id', member)
        self.assertIn(u'address', member)
        self.assertNotIn(u'name', member)
        self.assertNotIn(u'monitor_address', member)

    def test_get_all_filter(self):
        mem1 = self.create_member(self.pool_id,
                                  '192.0.2.1',
                                  80,
                                  name='member1').get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id,
                           '192.0.2.2',
                           80,
                           name='member2').get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.create_member(self.pool_id,
                           '192.0.2.3',
                           80,
                           name='member3').get(self.root_tag)
        self.set_lb_status(self.lb_id)

        members = self.get(self.members_path, params={
            'id': mem1['id']}).json
        self.assertEqual(1, len(members['members']))
        self.assertEqual(mem1['id'],
                         members['members'][0]['id'])

    def test_get_all_tags_filter(self):
        mem1 = self.create_member(
            self.pool_id,
            '192.0.2.1',
            80,
            name='member1',
            tags=['test_tag1', 'test_tag2']
        ).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        mem2 = self.create_member(
            self.pool_id,
            '192.0.2.2',
            80,
            name='member2',
            tags=['test_tag2', 'test_tag3']
        ).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        mem3 = self.create_member(
            self.pool_id,
            '192.0.2.3',
            80,
            name='member3',
            tags=['test_tag4', 'test_tag5']
        ).get(self.root_tag)
        self.set_lb_status(self.lb_id)

        mems = self.get(
            self.members_path,
            params={'tags': 'test_tag2'}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(2, len(mems))
        self.assertEqual(
            [mem1.get('id'), mem2.get('id')],
            [mem.get('id') for mem in mems]
        )

        mems = self.get(
            self.members_path,
            params={'tags': ['test_tag2', 'test_tag3']}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(1, len(mems))
        self.assertEqual(
            [mem2.get('id')],
            [mem.get('id') for mem in mems]
        )

        mems = self.get(
            self.members_path,
            params={'tags-any': 'test_tag2'}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(2, len(mems))
        self.assertEqual(
            [mem1.get('id'), mem2.get('id')],
            [mem.get('id') for mem in mems]
        )

        mems = self.get(
            self.members_path,
            params={'not-tags': 'test_tag2'}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(1, len(mems))
        self.assertEqual(
            [mem3.get('id')],
            [mem.get('id') for mem in mems]
        )

        mems = self.get(
            self.members_path,
            params={'not-tags-any': ['test_tag2', 'test_tag4']}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(0, len(mems))

        mems = self.get(
            self.members_path,
            params={'tags': 'test_tag2',
                    'tags-any': ['test_tag1', 'test_tag3']}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(2, len(mems))
        self.assertEqual(
            [mem1.get('id'), mem2.get('id')],
            [mem.get('id') for mem in mems]
        )

        mems = self.get(
            self.members_path,
            params={'tags': 'test_tag2', 'not-tags': 'test_tag2'}
        ).json.get(self.root_tag_list)
        self.assertIsInstance(mems, list)
        self.assertEqual(0, len(mems))

    def test_empty_get_all(self):
        response = self.get(self.members_path).json.get(self.root_tag_list)
        self.assertIsInstance(response, list)
        self.assertEqual(0, len(response))

    def test_create_sans_listener(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        self.assertEqual('192.0.2.1', api_member['address'])
        self.assertEqual(80, api_member['protocol_port'])
        self.assertIsNotNone(api_member['created_at'])
        self.assertIsNone(api_member['updated_at'])
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id,
            member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.ACTIVE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_CREATE,
            member_op_status=constants.NO_MONITOR)
        self.set_lb_status(self.lb_id)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id, member_id=api_member.get('id'))

    def test_create_authorized(self):
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)

        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': self.project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):

                api_member = self.create_member(
                    self.pool_id, '192.0.2.1', 80,
                    tags=['test_tag']).get(self.root_tag)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        self.assertEqual('192.0.2.1', api_member['address'])
        self.assertEqual(80, api_member['protocol_port'])
        self.assertEqual(['test_tag'], api_member['tags'])
        self.assertIsNotNone(api_member['created_at'])
        self.assertIsNone(api_member['updated_at'])
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id,
            member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.ACTIVE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_CREATE,
            member_op_status=constants.NO_MONITOR)
        self.set_lb_status(self.lb_id)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id, member_id=api_member.get('id'))

    def test_create_not_authorized(self):
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)

        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            api_member = self.create_member(
                self.pool_id, '192.0.2.1', 80, status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, api_member)

    def test_create_pool_in_error(self):
        project_id = uuidutils.generate_uuid()
        lb1 = self.create_load_balancer(uuidutils.generate_uuid(), name='lb1',
                                        project_id=project_id)
        lb1_id = lb1.get('loadbalancer').get('id')
        self.set_lb_status(lb1_id)
        pool1 = self.create_pool(
            lb1_id, constants.PROTOCOL_HTTP,
            constants.LB_ALGORITHM_ROUND_ROBIN).get('pool')
        pool1_id = pool1.get('id')
        self.set_lb_status(lb1_id)
        self.set_object_status(self.pool_repo, pool1_id,
                               provisioning_status=constants.ERROR)
        api_member = self.create_member(pool1_id, '192.0.2.1', 80, status=409)
        ref_msg = 'Pool %s is immutable and cannot be updated.' % pool1_id
        self.assertEqual(ref_msg, api_member.get('faultstring'))

    # TODO(rm_work) Remove after deprecation of project_id in POST (R series)
    def test_create_with_project_id_is_ignored(self):
        pid = uuidutils.generate_uuid()
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80, project_id=pid).get(self.root_tag)
        self.assertEqual(self.project_id, api_member['project_id'])

    def test_create_backup(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80, backup=True).get(self.root_tag)
        self.assertTrue(api_member['backup'])
        self.set_lb_status(self.lb_id)
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 81, backup=False).get(self.root_tag)
        self.assertFalse(api_member['backup'])

    def test_bad_create(self):
        member = {'name': 'test1'}
        self.post(self.members_path, self._build_body(member), status=400)

    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_create_with_bad_provider(self, mock_provider):
        mock_provider.side_effect = exceptions.ProviderDriverError(
            prov='bad_driver', user_msg='broken')
        response = self.create_member(self.pool_id, '192.0.2.1', 80,
                                      status=500)
        self.assertIn('Provider \'bad_driver\' reports error: broken',
                      response.get('faultstring'))

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_full_batch_members(self, mock_provider, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        member1 = {'address': '192.0.2.1', 'protocol_port': 80,
                   'project_id': self.project_id}
        member2 = {'address': '192.0.2.2', 'protocol_port': 80,
                   'project_id': self.project_id}
        member3 = {'address': '192.0.2.3', 'protocol_port': 80,
                   'project_id': self.project_id}
        member4 = {'address': '192.0.2.4', 'protocol_port': 80,
                   'project_id': self.project_id}
        member5 = {'address': '192.0.2.5', 'protocol_port': 80,
                   'project_id': self.project_id}
        member6 = {'address': '192.0.2.6', 'protocol_port': 80,
                   'project_id': self.project_id}
        members = [member1, member2, member3, member4]
        for m in members:
            self.create_member(pool_id=self.pool_id, **m)
            self.set_lb_status(self.lb_id)

        # We are only concerned about the batch update, so clear out the
        # create members calls above.
        mock_provider.reset_mock()

        req_dict = [member1, member2, member5, member6]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        expected_members = [
            ('192.0.2.1', 80, 'PENDING_UPDATE'),
            ('192.0.2.2', 80, 'PENDING_UPDATE'),
            ('192.0.2.3', 80, 'PENDING_DELETE'),
            ('192.0.2.4', 80, 'PENDING_DELETE'),
            ('192.0.2.5', 80, 'PENDING_CREATE'),
            ('192.0.2.6', 80, 'PENDING_CREATE'),
        ]

        provider_creates = []
        provider_updates = []
        for rm in returned_members:
            self.assertIn(
                (rm['address'],
                 rm['protocol_port'],
                 rm['provisioning_status']), expected_members)

            provider_dict = driver_utils.member_dict_to_provider_dict(rm)
            # Adjust for API response
            provider_dict['pool_id'] = self.pool_id
            if rm['provisioning_status'] == 'PENDING_UPDATE':
                del provider_dict['name']
                del provider_dict['subnet_id']
                provider_updates.append(driver_dm.Member(**provider_dict))
            elif rm['provisioning_status'] == 'PENDING_CREATE':
                provider_dict['name'] = None
                provider_creates.append(driver_dm.Member(**provider_dict))
        # Order matters here
        provider_creates += provider_updates

        mock_provider.assert_called_once_with(u'noop_driver',
                                              mock_driver.member_batch_update,
                                              self.pool_id, provider_creates)

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_create_batch_members(self, mock_provider, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        member5 = {'address': '192.0.2.5', 'protocol_port': 80,
                   'tags': ['test_tag1']}
        member6 = {'address': '192.0.2.6', 'protocol_port': 80,
                   'tags': ['test_tag2']}

        req_dict = [member5, member6]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        expected_members = [
            ('192.0.2.5', 80, 'PENDING_CREATE', ['test_tag1']),
            ('192.0.2.6', 80, 'PENDING_CREATE', ['test_tag2']),
        ]

        provider_members = []
        for rm in returned_members:
            self.assertIn(
                (rm['address'],
                 rm['protocol_port'],
                 rm['provisioning_status'],
                 rm['tags']), expected_members)

            provider_dict = driver_utils.member_dict_to_provider_dict(rm)
            # Adjust for API response
            provider_dict['pool_id'] = self.pool_id
            provider_dict['name'] = None
            provider_members.append(driver_dm.Member(**provider_dict))

        mock_provider.assert_called_once_with(u'noop_driver',
                                              mock_driver.member_batch_update,
                                              self.pool_id, provider_members)

    def test_create_batch_members_with_bad_subnet(self):
        subnet_id = uuidutils.generate_uuid()
        member5 = {'address': '10.0.0.5',
                   'protocol_port': 80,
                   'subnet_id': subnet_id}

        req_dict = [member5]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)

        with mock.patch(
                'octavia.common.utils.get_network_driver') as net_mock:
            net_mock.return_value.get_subnet = mock.Mock(
                side_effect=network_base.SubnetNotFound('Subnet not found'))
            response = self.put(path, body, status=400).json
            err_msg = 'Subnet ' + subnet_id + ' not found.'
            self.assertEqual(response.get('faultstring'), err_msg)

    def test_create_batch_members_with_invalid_address(self):
        # 169.254.169.254 is the default invalid member address
        member5 = {'address': '169.254.169.254',
                   'protocol_port': 80}

        req_dict = [member5]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)

        response = self.put(path, body, status=400).json
        err_msg = ("169.254.169.254 is not a valid option for member address")
        self.assertEqual(err_msg, response.get('faultstring'))

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_additive_only_batch_members(self, mock_provider, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        member1 = {'address': '192.0.2.1', 'protocol_port': 80}
        member2 = {'address': '192.0.2.2', 'protocol_port': 80}
        member3 = {'address': '192.0.2.3', 'protocol_port': 80}
        member4 = {'address': '192.0.2.4', 'protocol_port': 80}
        member5 = {'address': '192.0.2.5', 'protocol_port': 80}
        member6 = {'address': '192.0.2.6', 'protocol_port': 80}
        members = [member1, member2, member3, member4]
        for m in members:
            self.create_member(pool_id=self.pool_id, **m)
            self.set_lb_status(self.lb_id)

        # We are only concerned about the batch update, so clear out the
        # create members calls above.
        mock_provider.reset_mock()

        req_dict = [member1, member2, member5, member6]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        path = "{}?additive_only=True".format(path)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        # Members 1+2 should be updated, 3+4 left alone, and 5+6 created
        expected_members = [
            ('192.0.2.1', 80, 'PENDING_UPDATE'),
            ('192.0.2.2', 80, 'PENDING_UPDATE'),
            ('192.0.2.3', 80, 'ACTIVE'),
            ('192.0.2.4', 80, 'ACTIVE'),
            ('192.0.2.5', 80, 'PENDING_CREATE'),
            ('192.0.2.6', 80, 'PENDING_CREATE'),
        ]

        provider_creates = []
        provider_updates = []
        provider_ignored = []
        for rm in returned_members:
            self.assertIn(
                (rm['address'],
                 rm['protocol_port'],
                 rm['provisioning_status']), expected_members)

            provider_dict = driver_utils.member_dict_to_provider_dict(rm)
            # Adjust for API response
            provider_dict['pool_id'] = self.pool_id
            if rm['provisioning_status'] == 'PENDING_UPDATE':
                del provider_dict['name']
                del provider_dict['subnet_id']
                provider_updates.append(driver_dm.Member(**provider_dict))
            elif rm['provisioning_status'] == 'PENDING_CREATE':
                provider_dict['name'] = None
                provider_creates.append(driver_dm.Member(**provider_dict))
            elif rm['provisioning_status'] == 'ACTIVE':
                provider_dict['name'] = None
                provider_ignored.append(driver_dm.Member(**provider_dict))
        # Order matters here
        provider_creates += provider_updates
        provider_creates += provider_ignored

        mock_provider.assert_called_once_with(u'noop_driver',
                                              mock_driver.member_batch_update,
                                              self.pool_id, provider_creates)

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_update_batch_members(self, mock_provider, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        member1 = {'address': '192.0.2.1', 'protocol_port': 80,
                   'project_id': self.project_id}
        member2 = {'address': '192.0.2.2', 'protocol_port': 80,
                   'project_id': self.project_id}
        members = [member1, member2]
        for m in members:
            self.create_member(pool_id=self.pool_id, **m)
            self.set_lb_status(self.lb_id)

        # We are only concerned about the batch update, so clear out the
        # create members calls above.
        mock_provider.reset_mock()

        req_dict = [member1, member2]
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        expected_members = [
            ('192.0.2.1', 80, 'PENDING_UPDATE'),
            ('192.0.2.2', 80, 'PENDING_UPDATE'),
        ]

        provider_members = []
        for rm in returned_members:
            self.assertIn(
                (rm['address'],
                 rm['protocol_port'],
                 rm['provisioning_status']), expected_members)

            provider_dict = driver_utils.member_dict_to_provider_dict(rm)
            # Adjust for API response
            provider_dict['pool_id'] = self.pool_id
            del provider_dict['name']
            del provider_dict['subnet_id']
            provider_members.append(driver_dm.Member(**provider_dict))

        mock_provider.assert_called_once_with(u'noop_driver',
                                              mock_driver.member_batch_update,
                                              self.pool_id, provider_members)

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_delete_batch_members(self, mock_provider, mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        member3 = {'address': '192.0.2.3', 'protocol_port': 80}
        member4 = {'address': '192.0.2.4', 'protocol_port': 80}
        members = [member3, member4]
        for m in members:
            self.create_member(pool_id=self.pool_id, **m)
            self.set_lb_status(self.lb_id)

        # We are only concerned about the batch update, so clear out the
        # create members calls above.
        mock_provider.reset_mock()

        req_dict = []
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        expected_members = [
            ('192.0.2.3', 80, 'PENDING_DELETE'),
            ('192.0.2.4', 80, 'PENDING_DELETE'),
        ]

        provider_members = []
        for rm in returned_members:
            self.assertIn(
                (rm['address'],
                 rm['protocol_port'],
                 rm['provisioning_status']), expected_members)

        mock_provider.assert_called_once_with(u'noop_driver',
                                              mock_driver.member_batch_update,
                                              self.pool_id, provider_members)

    @mock.patch('octavia.api.drivers.driver_factory.get_driver')
    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_delete_batch_members_already_empty(self, mock_provider,
                                                mock_get_driver):
        mock_driver = mock.MagicMock()
        mock_driver.name = 'noop_driver'
        mock_get_driver.return_value = mock_driver

        req_dict = []
        body = {self.root_tag_list: req_dict}
        path = self.MEMBERS_PATH.format(pool_id=self.pool_id)
        self.put(path, body, status=202)
        returned_members = self.get(
            self.MEMBERS_PATH.format(pool_id=self.pool_id)
        ).json.get(self.root_tag_list)

        self.assertEqual([], returned_members)

        mock_provider.assert_not_called()

    def test_create_with_attached_listener(self):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.assertEqual('192.0.2.1', api_member['address'])
        self.assertEqual(80, api_member['protocol_port'])
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_CREATE,
            member_op_status=constants.NO_MONITOR)
        self.set_lb_status(self.lb_id)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'))

    def test_create_with_monitor_address_and_port(self):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            monitor_address='192.0.2.3',
            monitor_port=80).get(self.root_tag)
        self.assertEqual('192.0.2.1', api_member['address'])
        self.assertEqual(80, api_member['protocol_port'])
        self.assertEqual('192.0.2.3', api_member['monitor_address'])
        self.assertEqual(80, api_member['monitor_port'])
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_CREATE,
            member_op_status=constants.NO_MONITOR)
        self.set_lb_status(self.lb_id)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'))

    def test_create_with_health_monitor(self):
        self.create_health_monitor(self.pool_with_listener_id,
                                   constants.HEALTH_MONITOR_HTTP,
                                   1, 1, 1, 1)
        self.set_lb_status(self.lb_id)
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_CREATE,
            member_op_status=constants.OFFLINE)

    def test_duplicate_create(self):
        member = {'address': '192.0.2.1', 'protocol_port': 80,
                  'project_id': self.project_id}
        self.post(self.members_path, self._build_body(member))
        self.set_lb_status(self.lb_id)
        self.post(self.members_path, self._build_body(member), status=409)

    def test_create_with_bad_subnet(self):
        with mock.patch(
                'octavia.common.utils.get_network_driver') as net_mock:
            net_mock.return_value.get_subnet = mock.Mock(
                side_effect=network_base.SubnetNotFound('Subnet not found'))
            subnet_id = uuidutils.generate_uuid()
            response = self.create_member(self.pool_id, '192.0.2.1', 80,
                                          subnet_id=subnet_id, status=400)
            err_msg = 'Subnet ' + subnet_id + ' not found.'
            self.assertEqual(response.get('faultstring'), err_msg)

    def test_create_with_valid_subnet(self):
        with mock.patch(
                'octavia.common.utils.get_network_driver') as net_mock:
            subnet_id = uuidutils.generate_uuid()
            net_mock.return_value.get_subnet.return_value = subnet_id
            response = self.create_member(
                self.pool_id, '192.0.2.1', 80,
                subnet_id=subnet_id).get(self.root_tag)
            self.assertEqual('192.0.2.1', response['address'])
            self.assertEqual(80, response['protocol_port'])
            self.assertEqual(subnet_id, response['subnet_id'])

    def test_create_bad_port_number(self):
        member = {'address': '192.0.2.3',
                  'protocol_port': constants.MIN_PORT_NUMBER - 1}
        resp = self.post(self.members_path, self._build_body(member),
                         status=400)
        self.assertIn('Value should be greater or equal to',
                      resp.json.get('faultstring'))
        member = {'address': '192.0.2.3',
                  'protocol_port': constants.MAX_PORT_NUMBER + 1}
        resp = self.post(self.members_path, self._build_body(member),
                         status=400)
        self.assertIn('Value should be lower or equal to',
                      resp.json.get('faultstring'))

    def test_create_over_quota(self):
        self.start_quota_mock(data_models.Member)
        member = {'address': '192.0.2.3', 'protocol_port': 81}
        self.post(self.members_path, self._build_body(member), status=403)

    def test_update_with_attached_listener(self):
        old_name = "name1"
        new_name = "name2"
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            name=old_name).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        new_member = {'name': new_name}
        response = self.put(
            self.member_path_listener.format(member_id=api_member.get('id')),
            self._build_body(new_member)).json.get(self.root_tag)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_UPDATE)
        self.set_lb_status(self.lb_id)
        self.assertEqual(new_name, response.get('name'))
        self.assertEqual(api_member.get('created_at'),
                         response.get('created_at'))
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'))

    def test_update_authorized(self):
        old_name = "name1"
        new_name = "name2"
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            name=old_name, tags=['old_tag']).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        new_member = {'name': new_name, 'tags': ['new_tag']}
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': self.project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                member_path = self.member_path_listener.format(
                    member_id=api_member.get('id'))
                response = self.put(
                    member_path,
                    self._build_body(new_member)).json.get(self.root_tag)

        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_UPDATE)
        self.set_lb_status(self.lb_id)
        self.assertEqual(new_name, response.get('name'))
        self.assertEqual(['new_tag'], response['tags'])
        self.assertEqual(api_member.get('created_at'),
                         response.get('created_at'))
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'))

    def test_update_not_authorized(self):
        old_name = "name1"
        new_name = "name2"
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            name=old_name).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        new_member = {'name': new_name}
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            member_path = self.member_path_listener.format(
                member_id=api_member.get('id'))
            response = self.put(
                member_path,
                self._build_body(new_member), status=403)

        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, response.json)

        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=api_member.get('id'),
            lb_prov_status=constants.ACTIVE,
            listener_prov_status=constants.ACTIVE,
            pool_prov_status=constants.ACTIVE,
            member_prov_status=constants.ACTIVE)

    def test_update_sans_listener(self):
        old_name = "name1"
        new_name = "name2"
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80, name=old_name).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        member_path = self.member_path.format(
            member_id=api_member.get('id'))
        new_member = {'name': new_name}
        response = self.put(
            member_path, self._build_body(new_member)).json.get(self.root_tag)
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id, member_id=api_member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.ACTIVE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_UPDATE)
        self.set_lb_status(self.lb_id)
        self.assertEqual(new_name, response.get('name'))
        self.assertEqual(api_member.get('created_at'),
                         response.get('created_at'))
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_id, member_id=api_member.get('id'))

    def test_bad_update(self):
        api_member = self.create_member(
            self.pool_id, '192.0.2.1', 80).get(self.root_tag)
        new_member = {'protocol_port': 'ten'}
        self.put(self.member_path.format(member_id=api_member.get('id')),
                 self._build_body(new_member), status=400)

    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_update_with_bad_provider(self, mock_provider):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            name="member1").get(self.root_tag)
        self.set_lb_status(self.lb_id)
        new_member = {'name': "member2"}
        mock_provider.side_effect = exceptions.ProviderDriverError(
            prov='bad_driver', user_msg='broken')
        response = self.put(self.member_path_listener.format(
            member_id=api_member.get('id')), self._build_body(new_member),
            status=500)
        self.assertIn('Provider \'bad_driver\' reports error: broken',
                      response.json.get('faultstring'))

    def test_update_unset_defaults(self):
        old_name = "name1"
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80,
            name=old_name, backup=True, monitor_address='192.0.2.2',
            monitor_port=8888, weight=10).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        unset_params = {'name': None, 'backup': None, 'monitor_address': None,
                        'monitor_port': None, 'weight': None}
        member_path = self.member_path_listener.format(
            member_id=api_member.get('id'))
        response = self.put(member_path, self._build_body(unset_params))
        response = response.json.get(self.root_tag)

        self.assertFalse(response['backup'])
        self.assertIsNone(response['monitor_address'])
        self.assertIsNone(response['monitor_port'])
        self.assertEqual('', response['name'])
        self.assertEqual(constants.DEFAULT_WEIGHT, response['weight'])

    def test_delete(self):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id'))).json.get(self.root_tag)
        api_member['provisioning_status'] = constants.ACTIVE
        api_member['operating_status'] = constants.ONLINE
        self.assertIsNone(api_member.pop('updated_at'))
        self.assertIsNotNone(member.pop('updated_at'))
        self.assertEqual(api_member, member)
        self.delete(self.member_path_listener.format(
            member_id=api_member.get('id')))
        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_DELETE)

        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id')), status=404)

    def test_delete_authorized(self):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id'))).json.get(self.root_tag)
        api_member['provisioning_status'] = constants.ACTIVE
        api_member['operating_status'] = constants.ONLINE
        self.assertIsNone(api_member.pop('updated_at'))
        self.assertIsNotNone(member.pop('updated_at'))
        self.assertEqual(api_member, member)

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': False,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': self.project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                self.delete(self.member_path_listener.format(
                    member_id=api_member.get('id')))
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=member.get('id'),
            lb_prov_status=constants.PENDING_UPDATE,
            listener_prov_status=constants.PENDING_UPDATE,
            pool_prov_status=constants.PENDING_UPDATE,
            member_prov_status=constants.PENDING_DELETE)

        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id')), status=404)

    def test_delete_not_authorized(self):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id'))).json.get(self.root_tag)
        api_member['provisioning_status'] = constants.ACTIVE
        api_member['operating_status'] = constants.ONLINE
        self.assertIsNone(api_member.pop('updated_at'))
        self.assertIsNotNone(member.pop('updated_at'))
        self.assertEqual(api_member, member)

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               self.project_id):
            self.delete(self.member_path_listener.format(
                member_id=api_member.get('id')), status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

        self.assert_correct_status(
            lb_id=self.lb_id, listener_id=self.listener_id,
            pool_id=self.pool_with_listener_id, member_id=member.get('id'),
            lb_prov_status=constants.ACTIVE,
            listener_prov_status=constants.ACTIVE,
            pool_prov_status=constants.ACTIVE,
            member_prov_status=constants.ACTIVE)

    def test_bad_delete(self):
        self.delete(self.member_path.format(
            member_id=uuidutils.generate_uuid()), status=404)

    def test_delete_mismatch_pool(self):
        # Create a pool that will not have the member, but is valid.
        self.pool = self.create_pool(self.lb_id, constants.PROTOCOL_HTTP,
                                     constants.LB_ALGORITHM_ROUND_ROBIN)
        bad_pool_id = self.pool.get('pool').get('id')
        self.set_lb_status(self.lb_id)
        # Create a member on our reference pool
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        # Attempt to delete the member using the wrong pool in the path
        member_path = self.MEMBERS_PATH.format(
            pool_id=bad_pool_id) + '/' + api_member['id']
        result = self.delete(member_path, status=404).json
        ref_msg = 'Member %s not found.' % api_member['id']
        self.assertEqual(ref_msg, result.get('faultstring'))

    @mock.patch('octavia.api.drivers.utils.call_provider')
    def test_delete_with_bad_provider(self, mock_provider):
        api_member = self.create_member(
            self.pool_with_listener_id, '192.0.2.1', 80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        member = self.get(self.member_path_listener.format(
            member_id=api_member.get('id'))).json.get(self.root_tag)
        api_member['provisioning_status'] = constants.ACTIVE
        api_member['operating_status'] = constants.ONLINE
        self.assertIsNone(api_member.pop('updated_at'))
        self.assertIsNotNone(member.pop('updated_at'))
        self.assertEqual(api_member, member)

        mock_provider.side_effect = exceptions.ProviderDriverError(
            prov='bad_driver', user_msg='broken')
        self.delete(self.member_path_listener.format(
            member_id=api_member.get('id')), status=500)

    def test_create_when_lb_pending_update(self):
        self.create_member(self.pool_id, address="192.0.2.2",
                           protocol_port=80)
        self.set_lb_status(self.lb_id)
        self.put(self.LB_PATH.format(lb_id=self.lb_id),
                 body={'loadbalancer': {'name': 'test_name_change'}})
        member = {'address': '192.0.2.1', 'protocol_port': 80,
                  'project_id': self.project_id}
        self.post(self.members_path,
                  body=self._build_body(member),
                  status=409)

    def test_update_when_lb_pending_update(self):
        member = self.create_member(
            self.pool_id, address="192.0.2.1", protocol_port=80,
            name="member1").get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.put(self.LB_PATH.format(lb_id=self.lb_id),
                 body={'loadbalancer': {'name': 'test_name_change'}})
        self.put(
            self.member_path.format(member_id=member.get('id')),
            body=self._build_body({'name': "member2"}), status=409)

    def test_delete_when_lb_pending_update(self):
        member = self.create_member(
            self.pool_id, address="192.0.2.1",
            protocol_port=80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.put(self.LB_PATH.format(lb_id=self.lb_id),
                 body={'loadbalancer': {'name': 'test_name_change'}})
        self.delete(self.member_path.format(
            member_id=member.get('id')), status=409)

    def test_create_when_lb_pending_delete(self):
        self.create_member(self.pool_id, address="192.0.2.1",
                           protocol_port=80)
        self.set_lb_status(self.lb_id)
        self.delete(self.LB_PATH.format(lb_id=self.lb_id),
                    params={'cascade': "true"})
        member = {'address': '192.0.2.2', 'protocol_port': 88,
                  'project_id': self.project_id}
        self.post(self.members_path, body=self._build_body(member),
                  status=409)

    def test_update_when_lb_pending_delete(self):
        member = self.create_member(
            self.pool_id, address="192.0.2.1", protocol_port=80,
            name="member1").get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.delete(self.LB_PATH.format(lb_id=self.lb_id),
                    params={'cascade': "true"})
        self.put(self.member_path.format(member_id=member.get('id')),
                 body=self._build_body({'name': "member2"}), status=409)

    def test_update_when_deleted(self):
        member = self.create_member(
            self.pool_id, address="10.0.0.1",
            protocol_port=80).get(self.root_tag)
        self.set_lb_status(self.lb_id, status=constants.DELETED)
        self.put(self.member_path.format(member_id=member.get('id')),
                 body=self._build_body({'name': "member2"}), status=404)

    def test_delete_when_lb_pending_delete(self):
        member = self.create_member(
            self.pool_id, address="192.0.2.1",
            protocol_port=80).get(self.root_tag)
        self.set_lb_status(self.lb_id)
        self.delete(self.LB_PATH.format(lb_id=self.lb_id),
                    params={'cascade': "true"})
        self.delete(self.member_path.format(
            member_id=member.get('id')), status=409)

    def test_delete_already_deleted(self):
        member = self.create_member(
            self.pool_id, address="192.0.2.1",
            protocol_port=80).get(self.root_tag)
        self.set_lb_status(self.lb_id, status=constants.DELETED)
        self.delete(self.member_path.format(
            member_id=member.get('id')), status=404)



# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class Cert(object, metaclass=abc.ABCMeta):
    """Base class to represent all certificates."""

    @abc.abstractmethod
    def get_certificate(self):
        """Returns the certificate."""

    @abc.abstractmethod
    def get_intermediates(self):
        """Returns the intermediate certificates as a list."""

    @abc.abstractmethod
    def get_private_key(self):
        """Returns the private key for the certificate."""

    @abc.abstractmethod
    def get_private_key_passphrase(self):
        """Returns the passphrase for the private key."""



#    Copyright 2017 Walmart Stores Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add availability_zone table

Revision ID: c761c8a71579
Revises: e37941b010db
Create Date: 2019-11-11 18:53:15.428386

"""

from alembic import op
import sqlalchemy as sa

from octavia.common import constants

# revision identifiers, used by Alembic.
revision = 'c761c8a71579'
down_revision = 'e37941b010db'


def upgrade():
    azp_table = op.create_table(
        u'availability_zone_profile',
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=False),
        sa.Column(u'provider_name', sa.String(255), nullable=False),
        sa.Column(u'availability_zone_data', sa.String(4096), nullable=False),
        sa.PrimaryKeyConstraint(u'id'))

    op.bulk_insert(
        azp_table,
        [
            {'id': constants.NIL_UUID, 'name': 'DELETED-PLACEHOLDER',
             'provider_name': 'DELETED', 'availability_zone_data': '{}'},
        ]
    )

    az_table = op.create_table(
        u'availability_zone',
        sa.Column(u'name', sa.String(255), nullable=False),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'enabled', sa.Boolean(), nullable=False),
        sa.Column(u'availability_zone_profile_id', sa.String(36),
                  nullable=False),
        sa.ForeignKeyConstraint([u'availability_zone_profile_id'],
                                [u'availability_zone_profile.id'],
                                name=u'fk_az_az_profile_id'),
        sa.PrimaryKeyConstraint(u'name'),)

    op.bulk_insert(
        az_table,
        [
            {'name': constants.NIL_UUID,
             'description': 'Placeholder for DELETED LBs with DELETED '
                            'availability zones',
             'enabled': False,
             'availability_zone_profile_id': constants.NIL_UUID}
        ]
    )



#    Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
from unittest import mock

from mock import call

from octavia_lib.api.drivers import exceptions as driver_exceptions
from octavia_lib.common import constants as lib_consts
from oslo_utils import uuidutils

from octavia.api.drivers.driver_agent import driver_updater
from octavia.common import exceptions
import octavia.tests.unit.base as base


class TestDriverUpdater(base.TestCase):

    @mock.patch('octavia.db.repositories.LoadBalancerRepository')
    @mock.patch('octavia.db.repositories.ListenerRepository')
    @mock.patch('octavia.db.repositories.L7PolicyRepository')
    @mock.patch('octavia.db.repositories.L7RuleRepository')
    @mock.patch('octavia.db.repositories.PoolRepository')
    @mock.patch('octavia.db.repositories.HealthMonitorRepository')
    @mock.patch('octavia.db.repositories.MemberRepository')
    @mock.patch('octavia.db.api.get_session')
    def setUp(self, mock_get_session, mock_member_repo, mock_health_repo,
              mock_pool_repo, mock_l7r_repo, mock_l7p_repo, mock_list_repo,
              mock_lb_repo):
        super(TestDriverUpdater, self).setUp()
        self.mock_session = "FAKE_DB_SESSION"
        mock_get_session.return_value = self.mock_session

        member_mock = mock.MagicMock()
        mock_member_repo.return_value = member_mock
        self.mock_member_repo = member_mock
        health_mock = mock.MagicMock()
        mock_health_repo.return_value = health_mock
        self.mock_health_repo = health_mock
        pool_mock = mock.MagicMock()
        mock_pool_repo.return_value = pool_mock
        self.mock_pool_repo = pool_mock
        l7r_mock = mock.MagicMock()
        mock_l7r_repo.return_value = l7r_mock
        self.mock_l7r_repo = l7r_mock
        l7p_mock = mock.MagicMock()
        mock_l7p_repo.return_value = l7p_mock
        self.mock_l7p_repo = l7p_mock
        list_mock = mock.MagicMock()
        mock_list_repo.return_value = list_mock
        self.mock_list_repo = list_mock
        lb_mock = mock.MagicMock()
        mock_lb_repo.return_value = lb_mock
        self.mock_lb_repo = lb_mock
        self.driver_updater = driver_updater.DriverUpdater()
        self.ref_ok_response = {lib_consts.STATUS_CODE:
                                lib_consts.DRVR_STATUS_CODE_OK}
        mock_lb = mock.MagicMock()
        self.lb_id = uuidutils.generate_uuid()
        self.lb_project_id = uuidutils.generate_uuid()
        mock_lb.id = self.lb_id
        mock_lb.project_id = self.lb_project_id
        mock_lb.provisioning_status = lib_consts.ACTIVE
        self.lb_data_model = 'FakeLBDataModel'
        self.mock_lb_repo.model_class.__data_model__ = self.lb_data_model
        self.mock_lb_repo.get.return_value = mock_lb

    @mock.patch('octavia.common.utils.get_network_driver')
    def test_check_for_lb_vip_deallocate(self, mock_get_net_drvr):
        mock_repo = mock.MagicMock()
        mock_lb = mock.MagicMock()
        mock_vip = mock.MagicMock()
        mock_octavia_owned = mock.PropertyMock(side_effect=[True, False])
        type(mock_vip).octavia_owned = mock_octavia_owned
        mock_lb.vip = mock_vip
        mock_repo.get.return_value = mock_lb
        mock_net_drvr = mock.MagicMock()
        mock_get_net_drvr.return_value = mock_net_drvr

        self.driver_updater._check_for_lb_vip_deallocate(mock_repo, 'bogus_id')
        mock_net_drvr.deallocate_vip.assert_called_once_with(mock_vip)

        mock_net_drvr.reset_mock()
        self.driver_updater._check_for_lb_vip_deallocate(mock_repo, 'bogus_id')
        mock_net_drvr.deallocate_vip.assert_not_called()

    @mock.patch('octavia.db.repositories.Repositories.decrement_quota')
    @mock.patch('octavia.db.api.get_session')
    def test_decrement_quota(self, mock_get_session, mock_dec_quota):
        mock_session = mock.MagicMock()
        mock_get_session.return_value = mock_session
        mock_dec_quota.side_effect = [mock.DEFAULT,
                                      exceptions.OctaviaException('Boom')]

        self.driver_updater._decrement_quota(self.mock_lb_repo,
                                             'FakeName', self.lb_id)
        mock_dec_quota.assert_called_once_with(
            mock_session, self.mock_lb_repo.model_class.__data_model__,
            self.lb_project_id)
        mock_session.commit.assert_called_once()
        mock_session.rollback.assert_not_called()

        # Test exception path
        mock_dec_quota.reset_mock()
        mock_session.reset_mock()
        self.assertRaises(exceptions.OctaviaException,
                          self.driver_updater._decrement_quota,
                          self.mock_lb_repo, 'FakeName', self.lb_id)
        mock_dec_quota.assert_called_once_with(
            mock_session, self.mock_lb_repo.model_class.__data_model__,
            self.lb_project_id)
        mock_session.commit.assert_not_called()
        mock_session.rollback.assert_called_once()

        # Test already deleted path
        mock_dec_quota.reset_mock()
        mock_session.reset_mock()
        # Create a local mock LB and LB_repo for this test
        mock_lb = mock.MagicMock()
        mock_lb.id = self.lb_id
        mock_lb.provisioning_status = lib_consts.DELETED
        mock_lb_repo = mock.MagicMock()
        mock_lb_repo.model_class.__data_model__ = self.lb_data_model
        mock_lb_repo.get.return_value = mock_lb
        self.driver_updater._decrement_quota(mock_lb_repo,
                                             'FakeName', self.lb_id)
        mock_dec_quota.assert_not_called()
        mock_session.commit.assert_not_called()
        mock_session.rollback.assert_called_once()

    @mock.patch('octavia.api.drivers.driver_agent.driver_updater.'
                'DriverUpdater._decrement_quota')
    @mock.patch('octavia.api.drivers.driver_agent.driver_updater.'
                'DriverUpdater._check_for_lb_vip_deallocate')
    def test_process_status_update(self, mock_deallocate,
                                   mock_decrement_quota):
        mock_repo = mock.MagicMock()
        list_dict = {"id": 2,
                     lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                     lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        list_prov_dict = {"id": 2,
                          lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE}
        list_oper_dict = {"id": 2,
                          lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        list_deleted_dict = {
            "id": 2, lib_consts.PROVISIONING_STATUS: lib_consts.DELETED,
            lib_consts.OPERATING_STATUS: lib_consts.ONLINE}

        # Test with full record
        self.driver_updater._process_status_update(mock_repo, 'FakeName',
                                                   list_dict)
        mock_repo.update.assert_called_once_with(
            self.mock_session, 2, provisioning_status=lib_consts.ACTIVE,
            operating_status=lib_consts.ONLINE)
        mock_repo.delete.assert_not_called()

        # Test with only provisioning status record
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(mock_repo, 'FakeName',
                                                   list_prov_dict)
        mock_repo.update.assert_called_once_with(
            self.mock_session, 2, provisioning_status=lib_consts.ACTIVE)
        mock_repo.delete.assert_not_called()

        # Test with only operating status record
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(mock_repo, 'FakeName',
                                                   list_oper_dict)
        mock_repo.update.assert_called_once_with(
            self.mock_session, 2, operating_status=lib_consts.ONLINE)
        mock_repo.delete.assert_not_called()

        # Test with deleted but delete_record False
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(mock_repo, 'FakeName',
                                                   list_deleted_dict)
        mock_repo.update.assert_called_once_with(
            self.mock_session, 2, provisioning_status=lib_consts.DELETED,
            operating_status=lib_consts.ONLINE)
        mock_repo.delete.assert_not_called()
        mock_decrement_quota.assert_called_once_with(mock_repo, 'FakeName', 2)

        # Test with an empty update
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(mock_repo, 'FakeName',
                                                   {"id": 2})
        mock_repo.update.assert_not_called()
        mock_repo.delete.assert_not_called()

        # Test with deleted and delete_record True
        mock_decrement_quota.reset_mock()
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(
            mock_repo, 'FakeName', list_deleted_dict, delete_record=True)
        mock_repo.delete.assert_called_once_with(self.mock_session, id=2)
        mock_repo.update.assert_not_called()
        mock_decrement_quota.assert_called_once_with(mock_repo, 'FakeName', 2)

        # Test with LB Delete
        mock_decrement_quota.reset_mock()
        mock_repo.reset_mock()
        self.driver_updater._process_status_update(
            mock_repo, lib_consts.LOADBALANCERS, list_deleted_dict)
        mock_deallocate.assert_called_once_with(mock_repo, 2)
        mock_decrement_quota.assert_called_once_with(
            mock_repo, lib_consts.LOADBALANCERS, 2)

        # Test with an exception
        mock_repo.reset_mock()
        mock_repo.update.side_effect = Exception('boom')
        self.assertRaises(driver_exceptions.UpdateStatusError,
                          self.driver_updater._process_status_update,
                          mock_repo, 'FakeName', list_dict)

        # Test with no ID record
        mock_repo.reset_mock()
        self.assertRaises(driver_exceptions.UpdateStatusError,
                          self.driver_updater._process_status_update,
                          mock_repo, 'FakeName', {"fake": "data"})

    @mock.patch('octavia.api.drivers.driver_agent.driver_updater.'
                'DriverUpdater._process_status_update')
    def test_update_loadbalancer_status(self, mock_status_update):
        mock_status_update.side_effect = [
            mock.DEFAULT, mock.DEFAULT, mock.DEFAULT, mock.DEFAULT,
            mock.DEFAULT, mock.DEFAULT, mock.DEFAULT,
            driver_exceptions.UpdateStatusError(
                fault_string='boom', status_object='fruit',
                status_object_id='1', status_record='grape'),
            Exception('boom')]
        lb_dict = {"id": 1, lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                   lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        list_dict = {"id": 2,
                     lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                     lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        pool_dict = {"id": 3,
                     lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                     lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        member_dict = {"id": 4,
                       lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                       lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        hm_dict = {"id": 5, lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                   lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        l7p_dict = {"id": 6, lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                    lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        l7r_dict = {"id": 7, lib_consts.PROVISIONING_STATUS: lib_consts.ACTIVE,
                    lib_consts.OPERATING_STATUS: lib_consts.ONLINE}
        status_dict = {lib_consts.LOADBALANCERS: [lb_dict],
                       lib_consts.LISTENERS: [list_dict],
                       lib_consts.POOLS: [pool_dict],
                       lib_consts.MEMBERS: [member_dict],
                       lib_consts.HEALTHMONITORS: [hm_dict],
                       lib_consts.L7POLICIES: [l7p_dict],
                       lib_consts.L7RULES: [l7r_dict]}

        result = self.driver_updater.update_loadbalancer_status(
            copy.deepcopy(status_dict))

        calls = [call(self.mock_member_repo, lib_consts.MEMBERS, member_dict,
                      delete_record=True),
                 call(self.mock_health_repo, lib_consts.HEALTHMONITORS,
                      hm_dict, delete_record=True),
                 call(self.mock_pool_repo, lib_consts.POOLS, pool_dict,
                      delete_record=True),
                 call(self.mock_l7r_repo, lib_consts.L7RULES, l7r_dict,
                      delete_record=True),
                 call(self.mock_l7p_repo, lib_consts.L7POLICIES, l7p_dict,
                      delete_record=True),
                 call(self.mock_list_repo, lib_consts.LISTENERS, list_dict,
                      delete_record=True),
                 call(self.mock_lb_repo, lib_consts.LOADBALANCERS,
                      lb_dict)]
        mock_status_update.assert_has_calls(calls)
        self.assertEqual(self.ref_ok_response, result)

        # Test empty status updates
        mock_status_update.reset_mock()
        result = self.driver_updater.update_loadbalancer_status({})
        mock_status_update.assert_not_called()
        self.assertEqual(self.ref_ok_response, result)

        # Test UpdateStatusError case
        ref_update_status_error = {
            lib_consts.FAULT_STRING: 'boom',
            lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
            lib_consts.STATUS_OBJECT: 'fruit',
            lib_consts.STATUS_OBJECT_ID: '1'}
        result = self.driver_updater.update_loadbalancer_status(
            copy.deepcopy(status_dict))
        self.assertEqual(ref_update_status_error, result)

        # Test general exceptions
        result = self.driver_updater.update_loadbalancer_status(
            copy.deepcopy(status_dict))
        self.assertEqual({
            lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
            lib_consts.FAULT_STRING: 'boom'}, result)

    @mock.patch('octavia.db.repositories.ListenerStatisticsRepository.replace')
    def test_update_listener_statistics(self, mock_replace):
        listener_stats_list = [{"id": 1, "active_connections": 10,
                                         "bytes_in": 20,
                                         "bytes_out": 30,
                                         "request_errors": 40,
                                         "total_connections": 50},
                               {"id": 2, "active_connections": 60,
                                         "bytes_in": 70,
                                         "bytes_out": 80,
                                         "request_errors": 90,
                                         "total_connections": 100}]
        listener_stats_dict = {"listeners": listener_stats_list}

        mock_replace.side_effect = [mock.DEFAULT, mock.DEFAULT,
                                    Exception('boom')]
        result = self.driver_updater.update_listener_statistics(
            copy.deepcopy(listener_stats_dict))
        calls = [call(self.mock_session, 1, 1, active_connections=10,
                      bytes_in=20, bytes_out=30, request_errors=40,
                      total_connections=50),
                 call(self.mock_session, 2, 2, active_connections=60,
                      bytes_in=70, bytes_out=80, request_errors=90,
                      total_connections=100)]
        mock_replace.assert_has_calls(calls)
        self.assertEqual(self.ref_ok_response, result)

        # Test empty stats updates
        mock_replace.reset_mock()
        result = self.driver_updater.update_listener_statistics({})
        mock_replace.assert_not_called()
        self.assertEqual(self.ref_ok_response, result)

        # Test missing ID
        bad_id_dict = {"listeners": [{"notID": "one"}]}
        result = self.driver_updater.update_listener_statistics(bad_id_dict)
        ref_update_listener_stats_error = {
            lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
            lib_consts.STATS_OBJECT: lib_consts.LISTENERS,
            lib_consts.FAULT_STRING: "'id'"}
        self.assertEqual(ref_update_listener_stats_error, result)

        # Test for replace exception
        result = self.driver_updater.update_listener_statistics(
            copy.deepcopy(listener_stats_dict))
        ref_update_listener_stats_error = {
            lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
            lib_consts.STATS_OBJECT: lib_consts.LISTENERS,
            lib_consts.FAULT_STRING: 'boom', lib_consts.STATS_OBJECT_ID: 1}
        self.assertEqual(ref_update_listener_stats_error, result)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import collections


def create_iproute_ipv4_address(ip_address, broadcast_address, interface_name):
    """Returns a netlink/iproute (pyroute2) IPv4 address."""
    Stats = collections.namedtuple('Stats', ('qsize', 'delta', 'delay'))
    return (
        {'family': 2, 'prefixlen': 24, 'flags': 0, 'scope': 0, 'index': 2,
         'attrs': [('IFA_ADDRESS', ip_address), ('IFA_LOCAL', ip_address),
                   ('IFA_BROADCAST', broadcast_address),
                   ('IFA_LABEL', interface_name), ('IFA_FLAGS', 0),
                   ('IFA_CACHEINFO', {'ifa_preferred': 49256,
                                      'ifa_valid': 49256, 'cstamp': 1961,
                                      'tstamp': 73441020})],
         'header': {'length': 88, 'type': 20, 'flags': 2,
                    'sequence_number': 258, 'pid': 7590, 'error': None,
                    'stats': Stats(qsize=0, delta=0, delay=0)},
         'event': 'RTM_NEWADDR'},)


def create_iproute_ipv6_address(ip_address, interface_name):
    """Returns a netlink/iproute (pyroute2) IPv6 address."""
    Stats = collections.namedtuple('Stats', ('qsize', 'delta', 'delay'))
    return (
        {'family': 10, 'prefixlen': 64, 'flags': 0, 'scope': 0, 'index': 2,
         'attrs': [('IFA_CACHEINFO', {'ifa_preferred': 604503,
                                      'ifa_valid': 2591703, 'cstamp': 2038,
                                      'tstamp': 77073215}),
                   ('IFA_ADDRESS', '2001:db8:ffff:ffff:ffff:ffff:ffff:ffff'),
                   ('IFA_FLAGS', 768)],
         'header': {'length': 72, 'type': 20, 'flags': 2,
                    'sequence_number': 257, 'pid': 7590, 'error': None,
                    'stats': Stats(qsize=0, delta=0, delay=0)},
         'event': 'RTM_NEWADDR'},
        {'family': 10, 'prefixlen': 64, 'flags': 0, 'scope': 0, 'index': 2,
         'attrs': [('IFA_CACHEINFO', {'ifa_preferred': 604503,
                                      'ifa_valid': 2591703, 'cstamp': 2038,
                                      'tstamp': 77073215}),
                   ('IFA_ADDRESS', ip_address), ('IFA_FLAGS', 768)],
         'header': {'length': 72, 'type': 20, 'flags': 2,
                    'sequence_number': 257, 'pid': 7590, 'error': None,
                    'stats': Stats(qsize=0, delta=0, delay=0)},
         'event': 'RTM_NEWADDR'},)


def create_iproute_interface(interface_name):
    """Returns a netlink/iproute (pyroute2) interface."""
    Stats = collections.namedtuple('Stats', ('qsize', 'delta', 'delay'))
    return [{
        'family': 0, '__align': (), 'ifi_type': 1, 'index': 2, 'flags': 69699,
        'change': 0,
        'attrs': [('IFLA_TXQLEN', 1000), ('IFLA_IFNAME', interface_name),
                  ('IFLA_OPERSTATE', 'UP'), ('IFLA_LINKMODE', 0),
                  ('IFLA_MTU', 1500), ('IFLA_GROUP', 0),
                  ('IFLA_PROMISCUITY', 0), ('IFLA_NUM_TX_QUEUES', 1),
                  ('IFLA_GSO_MAX_SEGS', 65535),
                  ('IFLA_GSO_MAX_SIZE', 65536), ('IFLA_NUM_RX_QUEUES', 1),
                  ('IFLA_CARRIER', 1), ('IFLA_QDISC', 'fq_codel'),
                  ('IFLA_CARRIER_CHANGES', 2), ('IFLA_PROTO_DOWN', 0),
                  ('IFLA_CARRIER_UP_COUNT', 1),
                  ('IFLA_CARRIER_DOWN_COUNT', 1),
                  ('IFLA_MAP', {'mem_start': 0, 'mem_end': 0, 'base_addr': 0,
                                'irq': 0, 'dma': 0, 'port': 0}),
                  ('IFLA_ADDRESS', '52:54:00:cf:37:9e'),
                  ('IFLA_BROADCAST', 'ff:ff:ff:ff:ff:ff'),
                  ('IFLA_STATS64', {
                      'rx_packets': 756091, 'tx_packets': 780292,
                      'rx_bytes': 234846748, 'tx_bytes': 208583687,
                      'rx_errors': 0, 'tx_errors': 0, 'rx_dropped': 0,
                      'tx_dropped': 0, 'multicast': 0, 'collisions': 0,
                      'rx_length_errors': 0, 'rx_over_errors': 0,
                      'rx_crc_errors': 0, 'rx_frame_errors': 0,
                      'rx_fifo_errors': 0, 'rx_missed_errors': 0,
                      'tx_aborted_errors': 0, 'tx_carrier_errors': 0,
                      'tx_fifo_errors': 0, 'tx_heartbeat_errors': 0,
                      'tx_window_errors': 0, 'rx_compressed': 0,
                      'tx_compressed': 0}),
                  ('IFLA_STATS', {
                      'rx_packets': 756091, 'tx_packets': 780292,
                      'rx_bytes': 234846748, 'tx_bytes': 208583687,
                      'rx_errors': 0, 'tx_errors': 0, 'rx_dropped': 0,
                      'tx_dropped': 0, 'multicast': 0, 'collisions': 0,
                      'rx_length_errors': 0, 'rx_over_errors': 0,
                      'rx_crc_errors': 0, 'rx_frame_errors': 0,
                      'rx_fifo_errors': 0, 'rx_missed_errors': 0,
                      'tx_aborted_errors': 0, 'tx_carrier_errors': 0,
                      'tx_fifo_errors': 0, 'tx_heartbeat_errors': 0,
                      'tx_window_errors': 0, 'rx_compressed': 0,
                      'tx_compressed': 0}),
                  ('IFLA_XDP', '05:00:02:00:00:00:00:00'),
                  ('IFLA_AF_SPEC', {
                      'attrs': [
                          ('AF_INET', {
                              'dummy': 65664, 'forwarding': 1,
                              'mc_forwarding': 0, 'proxy_arp': 0,
                              'accept_redirects': 1,
                              'secure_redirects': 1,
                              'send_redirects': 1, 'shared_media': 1,
                              'rp_filter': 1, 'accept_source_route': 1,
                              'bootp_relay': 0, 'log_martians': 0,
                              'tag': 0, 'arpfilter': 0, 'medium_id': 0,
                              'noxfrm': 0, 'nopolicy': 0,
                              'force_igmp_version': 0, 'arp_announce': 0,
                              'arp_ignore': 0, 'promote_secondaries': 0,
                              'arp_accept': 0, 'arp_notify': 0,
                              'accept_local': 0, 'src_vmark': 0,
                              'proxy_arp_pvlan': 0, 'route_localnet': 0,
                              'igmpv2_unsolicited_report_interval': 10000,
                              'igmpv3_unsolicited_report_interval': 1000}),
                          ('AF_INET6', {
                              'attrs': [('IFLA_INET6_FLAGS', 2147483648),
                                        ('IFLA_INET6_CACHEINFO', {
                                            'max_reasm_len': 65535,
                                            'tstamp': 1859,
                                            'reachable_time': 30708,
                                            'retrans_time': 1000}),
                                        ('IFLA_INET6_CONF', {
                                            'forwarding': 1, 'hop_limit': 64,
                                            'mtu': 1500, 'accept_ra': 2,
                                            'accept_redirects': 1,
                                            'autoconf': 1,
                                            'dad_transmits': 1,
                                            'router_solicitations': 4294967295,
                                            'router_solicitation_interval':
                                                4000,
                                            'router_solicitation_delay': 1000,
                                            'use_tempaddr': 0,
                                            'temp_valid_lft': 604800,
                                            'temp_preferred_lft': 86400,
                                            'regen_max_retry': 3,
                                            'max_desync_factor': 600,
                                            'max_addresses': 16,
                                            'force_mld_version': 0,
                                            'accept_ra_defrtr': 1,
                                            'accept_ra_pinfo': 1,
                                            'accept_ra_rtr_pref': 1,
                                            'router_probe_interval': 60000,
                                            'accept_ra_rt_info_max_plen': 0,
                                            'proxy_ndp': 0,
                                            'optimistic_dad': 0,
                                            'accept_source_route': 0,
                                            'mc_forwarding': 0,
                                            'disable_ipv6': 0,
                                            'accept_dad': 1, 'force_tllao': 0,
                                            'ndisc_notify': 0}),
                                        ('IFLA_INET6_STATS', {
                                            'num': 37, 'inpkts': 57817,
                                            'inoctets': 144065857,
                                            'indelivers': 36758,
                                            'outforwdatagrams': 0,
                                            'outpkts': 35062,
                                            'outoctets': 4796485,
                                            'inhdrerrors': 0,
                                            'intoobigerrors': 0,
                                            'innoroutes': 0, 'inaddrerrors': 0,
                                            'inunknownprotos': 0,
                                            'intruncatedpkts': 0,
                                            'indiscards': 0,
                                            'outdiscards': 0, 'outnoroutes': 0,
                                            'reasmtimeout': 0, 'reasmreqds': 0,
                                            'reasmoks': 0, 'reasmfails': 0,
                                            'fragoks': 0, 'fragfails': 0,
                                            'fragcreates': 0,
                                            'inmcastpkts': 23214,
                                            'outmcastpkts': 6546,
                                            'inbcastpkts': 0,
                                            'outbcastpkts': 0,
                                            'inmcastoctets': 2255059,
                                            'outmcastoctets': 589090,
                                            'inbcastoctets': 0,
                                            'outbcastoctets': 0,
                                            'csumerrors': 0,
                                            'noectpkts': 57860,
                                            'ect1pkts': 0, 'ect0pkts': 0,
                                            'cepkts': 0}),
                                        ('IFLA_INET6_ICMP6STATS', {
                                            'num': 6, 'inmsgs': 2337,
                                            'inerrors': 0, 'outmsgs': 176,
                                            'outerrors': 0, 'csumerrors': 0}),
                                        ('IFLA_INET6_TOKEN', '::'),
                                        ('IFLA_INET6_ADDR_GEN_MODE', 0)]})]})],
        'header': {'length': 1304, 'type': 16, 'flags': 0,
                   'sequence_number': 261, 'pid': 7590, 'error': None,
                   'stats': Stats(qsize=0, delta=0, delay=0)},
        'state': 'up', 'event': 'RTM_NEWLINK'}]
